/* Module DIALPRI
   Contains functions for the dialogue to e.g. select driver for a printer
   port */

#include <cgui.h>
#include <cgui/mem.h>
#include <allegro.h>

#include "spooler.h"
#include "dialpri.h"
#include "printini.h"
#include "cpri.ht"
#include "labldata.h"
#include "icondata.h"

typedef struct t_dialpri {
   t_printerq upd;              /* copy of q to edit */
   t_printerq *q;               /* pointer to real q-data */
   t_printerq backup;           /* copy of q to use at cancel. This is
                                   necessary */
   /* because the user expects the print-test function */
   /* to use the displayed settings, and the spooler */
   /* will use the real q-data, so at test-print-command */
   /* must copy the edited data to q (and can not resore */
   /* it since it don't know when the spooler is ready). */
   int id1, id2, id3;
} t_dialpri;

static const char *const*ptxt;
static BITMAP *gray_im;

static void CancelEdPriPort(t_dialpri * dialpri)
{
   Release(dialpri);
   CloseWin(NULL);
   StopProcessEvents();
}

static void ReInstall(t_printerq * q)
{
   if (q->installed) {
      _CpriSetPrinterDriver(q);
      if (q->vtable)
         q->drv = q->vtable->Init(q);
   }
}

static void DeInstall(t_printerq *q)
{
   _CpriCloseQueue(q);
}

static void QuitEdPriPort(void *data)
{
   t_dialpri *dialpri = data, *d;
   int i;

   for (i = 0, d = dialpri; i < 5; i++, d++) {
      DeInstall(d->q);          /* might have been re-installed by print-test
                                 */
      *d->q = d->backup;
      ReInstall(d->q);
   }
   CancelEdPriPort(dialpri);
}

static void ExitEdPriPort(void *data)
{
   t_dialpri *dialpri = data, *d;
   int i;

   for (i = 0, d = dialpri; i < 5; i++, d++) {
      DeInstall(d->q);          /* might have been re-installed by print-test
                                 */
      *d->q = d->upd;
      ReInstall(d->q);
   }
   _CpriSavePrinterSettings();
   CancelEdPriPort(dialpri);
}

static void PrintTest(void *data)
{
   t_dialpri *dialpri = data;
   t_printerq *q;

   q = dialpri->q;
   ReInstall(q);
//   if (q->PrintTest) ###
//      q->PrintTest(q);
}

static void SwitchState(void *data)
{
   t_dialpri *dialpri = data;

   if (dialpri->upd.installed) {
      Activate(dialpri->id1);
      Activate(dialpri->id2);
      Activate(dialpri->id3);
   } else {
      DeActivate(dialpri->id1);
      DeActivate(dialpri->id2);
      DeActivate(dialpri->id3);
   }
   Refresh(dialpri->id1);
   Refresh(dialpri->id2);
   Refresh(dialpri->id3);
}

static void Button4(char *tag, t_dialpri * dialpri)
{
   static const char *prilang_sel[NODRIVERS];
   int id;
   static const char * const symbol_sets[] = { "Windows", "DOS (PC-8)", "Roman-8" };
   int i;
   for (i = 0; i < NODRIVERS; i++)
#ifdef ALLEGRO_DOS
      prilang_sel[i] = ptxt[pt_pribrands + i];
#elif defined ALLEGRO_WINDOWS
      prilang_sel[i] = "Windows";
#else
      prilang_sel[i] = "Linux";
#endif
   dialpri->id3 = AddButton(DOWNLEFT | ALIGNRIGHT, ptxt[pt_print_test], PrintTest, dialpri);
   DeActivate(dialpri->id3);
   dialpri->id2 = AddDropDownS(LEFT, 0, ptxt[pt_symbol_set],
                               &dialpri->upd.dostext, symbol_sets, 3);
   dialpri->id1 = AddDropDownS(LEFT, 0, ptxt[PRINTER_BRAND],
                               &dialpri->upd.brand, prilang_sel, NODRIVERS);
   id = AddCheck(LEFT, ptxt[pt_connected], &dialpri->upd.installed);
   AddHandler(id, SwitchState, dialpri);
   AddTag(LEFT, tag);
   /* SetUserHandler(dialpri->id1, SwitchBrand, dialpri); */
   if (!dialpri->upd.installed) {
      DeActivate(dialpri->id1);
      DeActivate(dialpri->id2);
      DeActivate(dialpri->id3);
   }
   /* else if (dialpri->upd.ds.brand!=BRAND_UNKNOWN)
      DeActivate(dialpri->id2); */
}

extern const char *const*CpriGetDriverNames(void)
{
   int nr;

   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      exit(-1);
   return ptxt + pt_pribrands;
}

extern const char *const*CpriGetQueueNames(void)
{
   int nr;

   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      exit(-1);
   return ptxt + pt_selq;
}

extern void _CpriDestroyBarIcon(void)
{
   if (gray_im)
      Release(gray_im);
}

#include <allegro/internal/aintern.h>
static void InitCompiledBitmap(BITMAP *bmp)
{
   if ((unsigned)bmp->vtable >= 32)
      return;
   bmp->vtable = _get_vtable((int)bmp->vtable);
   bmp->write_bank = _stub_bank_switch;
   bmp->read_bank = _stub_bank_switch;
   bmp->seg = _default_ds();
}

static const char *bi_names[BAR_PATTERN_NR] = {"#backdashed", "#dashed", "#hlines", "#squares",
                        "#vlines", "#xsquares", "#gray"};
extern void _CpriInitBarIcons(void)
{
   int i;
   static int done;

   if (!done) {
      done = 1;
      ASSERT(BAR_PATTERN_NR==COUNT+1);
      set_color_conversion(COLORCONV_KEEP_TRANS|COLORCONV_TOTAL);
      for (i=0; i<COUNT; i++)
         InitCompiledBitmap(icons[i].dat);
      for (i=0; i<COUNT; i++)
         RegisterImage(icons[i].dat, bi_names[i] + 1, IMAGE_TRANS_BMP, 0);
   }
}

extern const char *const*CpriGetBarIcons(void)
{
   static int done;

   if (!done) {
      done = 1;
      _CpriInitBarIcons();
      gray_im = create_bitmap(18, 16);
      clear_to_color(gray_im, makecol(224,224,224));
      RegisterImage(gray_im, bi_names[BAR_PATTERN_NR-1]+1, IMAGE_TRANS_BMP, 0);
   }
   return bi_names;
}

extern void CpriPrinterPortDialogue(void)
{
   t_dialpri *dialpri, *d;
   int i, nr;

   if (_CpriIsPrinting2())
      return;
   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      exit(-1);
   CpriInitPrinter();
   dialpri = GetMem0(t_dialpri, NR_OF_QUEUES);
   for (i = 0, d = dialpri; i < 5; i++, d++) {
      DeInstall(printerqs + i);
      d->upd = printerqs[i];
      d->backup = printerqs[i];
      d->q = printerqs + i;
   }

   MkDialogue(ADAPTIVE, ptxt[pt_adapt_to_connected_printer], 0);
   AutoHotKeys(0);
   SetDistance(4, 10);

   StartContainer(DOWNLEFT | FILLSPACE | HORIZONTAL, ADAPTIVE,
                  ptxt[pt_direct_port_printing], CT_BORDER);
   SetDistance(8, 4);
   Button4("LPT1:", dialpri + QUEUE_LPT1);
   Button4("LPT2:", dialpri + QUEUE_LPT2);
   Button4("LPT3:", dialpri + QUEUE_LPT3);
   EndContainer();

   /* Dos-standard */
   StartContainer(DOWNLEFT | FILLSPACE | HORIZONTAL, ADAPTIVE,
                  ptxt[pt_selq + QUEUE_DOSPRINT], CT_BORDER);
   Button4("", dialpri + QUEUE_DOSPRINT);
   EndContainer();

   /* Network printer */
   StartContainer(DOWNLEFT | FILLSPACE | HORIZONTAL, ADAPTIVE,
                  ptxt[pt_net], CT_BORDER);
   Button4("", dialpri + QUEUE_NETPORT);
   AddTag(DOWNLEFT, ptxt[pt_procedure_name_for_net_printing]);
   AddEditBox(RIGHT, 160, "", FSTRING, MAXNETCOMMAND,
              dialpri[QUEUE_NETPORT].upd.net1);
   AddEditBox(RIGHT, 160, "", FSTRING, MAXNETCOMMAND,
              dialpri[QUEUE_NETPORT].upd.net2);
   EndContainer();

   AddButton(DOWNLEFT, ptxt[pt_Cancel], QuitEdPriPort, dialpri);
   AddButton(DOWNLEFT, ptxt[pt_OK], ExitEdPriPort, dialpri);
   DisplayWin();
   ProcessEvents();
}
