/* Module DRV_CANO.C
   Contains the print-driver for Canon printer */
#ifdef DJGPP
#include <string.h>

#include <cgui.h>
#include <cgui/mem.h>

#include "spooler.h"
#include "drv_cano.h"
#include "cpri.h"
#include "pridrive.h"
#include "drv_common.h"

static char **drv_fmt;

static char **init_CANON_codes(void)
{
   int i;
   drv_fmt = GetMem(char *, 256);

   for (i = 0; i < 256; i++)
      drv_fmt[i] = "";
/* "\33\1330 K"; // symbol set */
   drv_fmt[(int)EPC_PORTRAIT] = "\33\13311h\33\1337 I\33\1330p";
   drv_fmt[(int)EPC_LANDSCAPE] = "\33\1331p";
   drv_fmt[(int)EPC_MANUALFEED] = "\33\1331q";
   drv_fmt[(int)EPC_CASSETTEFEED] = "\33\1330q";
   /* drv_fmt[EPC_BOLD]="\33\1331m"; drv_fmt[EPC_UNBOLD]="\33\13322m";
   drv_fmt[EPC_NSPACE] = "\33\13388j",
   drv_fmt[EPC_WBAR] = "\263\33\13396j\263\33\13396j\263";
   drv_fmt[EPC_WSPACE] = "\33\13381j"; */
   drv_fmt[(int)EPC_INDEXUPRIGHT] = "\33\13320k";
   drv_fmt[(int)EPC_INDEXWIDTH] = "\33\13320e";

   drv_fmt[(int)EPC_ABSMOVEX] = "\33[%dd";
   drv_fmt[(int)EPC_ABSMOVEY] = "\33[%d`";

   /* These are commands internally used by the dirver EPC_SETTAB
      EPC_TABTONEXT EPC_CLEARTABS EPC_AUTOGRID EPC_RIGHTALIGN EPC_HTMLTAG
      EPC_HTMLTAGEND EPC_DRAWTOPFRAME EPC_DRAWMIDFRAME EPC_DRAWBOTTTOMFRAME
      EPC_LINETHROUGH */
   return drv_fmt;
}

typedef struct t_driver {
   char *format;                /* A pointer to a format-string or other
                                   string which shall be sent to the printer
                                   before next character from the file */
   char *tmpstr;                /* A string that may be used by the driver to
                                   to host a printer command that are not
                                   predefined (that is generated during
                                   spooling) */
   t_tab_frame tf;              /* A list of the current tabs (in three
                                   different scalings) */
   t_tab margin;                /* Specifies the current margins. */
   int topmargin;
   int marginstate;             /* indicates that the top margin has been set
                                   but no newline has been encountered. If a
                                   fontsize commad is received while this
                                   flag is set, the y-position will be
                                   adjusted to keep the top margin the same. */
   int rowheight;               /* The heigth of each line used for demand
                                   line feed */
   int hightpercent;            /* The row-heigth expressed in % of standard
                                   height 4.2mm/12dpi font */
   int linepospercent;          /* The distance to the horizontal line from
                                   the line above in % of the rowheight */
   int ypos;                    /* current printing y-position */
   int bottom;                  /* "soft" bottom margin */
   int text2grid;               /* distance from text to the grid-border
                                   below */
   int gridon;                  /* It is necessary to set gridon after
                                   defining the set of tabs to be used */
   int indexwidth;
   int size;                    /* current font size */
   int style;                   /* current font style */
   int stroke;                  /* current stroke weight of font */
   int face;                    /* current font typeface */
   int macroplaying;
   int ishdr;
   int hdrid;                   /* macro-id of the header-macro */
   int headend;                 /* the end-position (y) of the header-macro */
   int istblhdr;
   int tblhdrid;                /* macro-id of the tableheader-macro */
   int tblheadend;              /* the end-position (y) of the
                                   tableheader-macro */
   int ul;                      /* underline indicator */
   int dos;                     /* flag indicating if character codes are to
                                   be converted to the PC-8 character set */
} t_driver;

extern void CanonInit(t_printerq * q)
{
   t_driver *dd;

   dd = q->drv;
   if (drv_fmt == NULL)
      drv_fmt = init_CANON_codes();
   dd = q->drv;
   if (dd == NULL)
      dd = q->drv = GetMem0(t_driver, 1);
   dd->size = 12;
   dd->face = COURIER;
   dd->size = 12;
   dd->tmpstr = ResizeMem(char, dd->tmpstr, 1000);

   dd->format = "";
}

extern void CanonClose(t_printerq * q)
{
   if (q->drv) {
      if (!_CpriEmptyQ(q))
         _CpriFlushPQ(q);
      if (q->drv->tmpstr)
         Release(q->drv->tmpstr);
      if (q->drv->tf.tabs)
         Release(q->drv->tf.tabs);
      Release(q->drv);
      q->drv = NULL;
   }
}

extern void CanonTest(t_printerq * q nouse)
{
}

extern int CanonDriver(t_printerq * q)
{
   int c;
   int param;
   t_driver *dd;
   char s[500];

   dd = q->drv;
   if (*dd->format)
      return *dd->format++;
   c = getc(q->fil);
   q->bytes--;
   switch (c) {
   case 0x33:
      c = getc(q->fil);
      q->bytes--;
      switch (c) {
      case EPC_PRINT_CHECK:
      case EPC_END_START_SEQUENCE:
         return CanonDriver(q);
         break;
      case EPC_SETSCALE:
         _CpriGetParam(q);
         return CanonDriver(q);
      case EPC_NEWPAGE:
         dd->format = dd->tmpstr;
         sprintf(dd->format, "\14\15");
         return *dd->format++;
         break;
      case EPC_PRINTBAR:
         _CpriGetParam(q);
         _CpriGetParam(q);
         _CpriGetParam(q);
         _CpriGetParam(q);
         _CpriGetParam(q);
         return CanonDriver(q);
         break;
      case EPC_PRINTBARCODE:
      case EPC_INDEXUPRIGHT:
      case EPC_INDEXUPLEFT:
      case EPC_INDEXUPRIGHTB:
      case EPC_INDEXUPLEFTB:
      case EPC_NOCOPIES:
         _CpriGetSubsequentText(q, s, dd->dos);
         return CanonDriver(q);
         break;
      case EPC_ACTIVATETABS:
      case EPC_SINGLETABLEROW: /* print a single row of a table */
         return CanonDriver(q);
         break;
      case EPC_ALIGN:
         _CpriGetParam(q);         /* dummy reads */
         _CpriGetSubsequentText(q, s, dd->dos);
         return CanonDriver(q);
         break;
      case EPC_TYPEFACES:       /* commands that contains 2 paramters */
         _CpriGetParam(q);
         return CanonDriver(q);
      case EPC_RESET:
         dd->dos = 0;
         dd->format = dd->tmpstr;
         sprintf(dd->format, "\33<");
         return *dd->format++;
      case EPC_DRAWRECT:        /* n.u. */
         _CpriGetParam(q);
         _CpriGetParam(q);
         return CanonDriver(q);
         break;
      case EPC_RELMOVEX:
      case EPC_RELMOVEY:
         param = _CpriGetParam(q);
         return CanonDriver(q);
         break;
      case EPC_ABSMOVEX:        /* commands that contains 1 paramter */
      case EPC_ABSMOVEY:
      case EPC_MACROID:
         param = _CpriGetParam(q);
         dd->format = dd->tmpstr;
         sprintf(dd->format, drv_fmt[c], param);
         return *dd->format++;
      case EPC_NEWLINE:
         dd->tf.curtab = 0;
         return '\n';
         break;
      case EPC_DOSTEXT:
         dd->dos = _CpriGetParam(q);
         return CanonDriver(q);
         break;
      case EPC_SIZE:
         param = _CpriGetParam(q);
         dd->format = dd->tmpstr;
         sprintf(dd->format, "\33\133%d;%d B", 100 * param / 12,
                 100 * param / 12);
         return *dd->format++;
         break;
      case EPC_SETTAB:          /* settab */
         _CpriSetTab(&dd->tf, &dd->margin, _CpriGetParam(q), 300);
         return CanonDriver(q);
         break;
      case EPC_TABTONEXT:       /* tab to next */
         if (dd->tf.curtab >= dd->tf.n)
            return CanonDriver(q);
         dd->format = dd->tmpstr;
         sprintf(dd->format, absmovx, dd->tf.tabs[dd->tf.curtab]);
         dd->tf.curtab++;
         return *dd->format++;
         break;
      case EPC_CLEARTABS:       /* clear tabs */
         dd->tf.n = dd->gridon = 0;
         if (dd->tf.tabs)
            Release(dd->tf.tabs);
         dd->tf.tabs = NULL;
         return CanonDriver(q);
         break;
      case EPC_STARTTABLE:   /* set auto grid */
         return CanonDriver(q);
         break;
      case EPC_HTMLTAG:      /* html-tag - clear input and ignore */
      case EPC_DEACTIVATETAB:
         param = _CpriGetParam(q);
         return CanonDriver(q);
         break;
      case EPC_HTMLTAGEND:   /* html-tagend  - ignore */
         return CanonDriver(q);
         break;
      default:
         dd->format = drv_fmt[c];
         return *dd->format++;
         break;
      }
      break;
   case '\n':
      dd->tf.curtab = 0;
      return c;
      break;
   case EOF:
      if (_CpriEmptyQ(q))
         return EOF;
      return CanonDriver(q);
      break;
   default:
      return c;
      break;
   }
}
#endif
