/* Module DRV_COMMON.C
   Contains the common functions for printer drivers */

#include <string.h>
#include <cgui/mem.h>

#include "cpri.h"
#include "spooler.h"
#include "pridrive.h"

#include "drv_common.h"
#include "width.h"
#include "dialpri.h"
#include "cpri.ht"


static const char * const typeface_names[] = {
   "Flareserif (Albertus)",
   "Antique Olive",
   "Times",
   "Univers (Arial)",
   "Univers Condensed (Arial Narrow)",
   "Clarendon Condensed (Coronet)",
   /* Propotional fonts */
   "Courier (fix)",
   "Gothic (fix)",
   "Script",
   "FONT 10",
   "FONT 11"
};

extern void _CpriMoveToNextCell(t_tab_frame * tf)
{
   tf->prevtab=tf->curtab;
   do
      tf->curtab++;
   while (tf->curtab < tf->n && tf->tabs[tf->curtab].inactive);
}

extern void _CpriRemovePriTab(t_tab_frame * tf, unsigned tabno)
{
   tf->tabs[tabno].inactive = 1;
}

extern void _CpriMergeAllColumns(t_tab_frame *tf)
{
   int i;
   for (i=1; i<tf->n-1; i++)
       _CpriRemovePriTab(tf, i);
}
   
extern void _CpriActivateTabs(t_tab_frame * tf)
{
   int i;

   for (i = 0; i < tf->n; i++)
      tf->tabs[i].inactive = 0;
}

static void SetNextTab(t_tab_frame *tf, int nextpos, float scaling)
{
   if (tf->maxtabs < tf->n + 1) {
      tf->maxtabs += 20;
      tf->tabs = ResizeMem(t_tab, tf->tabs, tf->maxtabs);
   }
   tf->tabs[tf->n].inactive = 0;
   tf->tabs[tf->n].mm = nextpos;
   tf->tabs[tf->n].scaled = (int)(scaling * nextpos + 0.5);
   tf->tabs[tf->n].block = (int)(nextpos / 2.54 + 0.5);
   tf->n++;
}

extern void _CpriSetTab(t_tab_frame *tf, int w, float scaling)
{
   SetNextTab(tf, tf->tabs[tf->n - 1].mm + w, scaling);
}

extern void _CpriInitTabs(t_tab_frame *tf, int leftx, float scaling)
{
   tf->n = 0;
   SetNextTab(tf, leftx, scaling);
}

extern void _CpriDestroyTabFrame(t_tab_frame *f)
{
   Release(f->tabs);
   f->tabs = NULL;
   f->maxtabs = 0;
   f->n = 0;
}

extern void _CpriCopyTabFrame(t_tab_frame *from, t_tab_frame *to)
{
   *to = *from;
/*   if(to->tabs)
	   Release(to->tabs);*/
   to->tabs = GetMem0(t_tab, to->maxtabs);
   memmove(to->tabs, from->tabs, sizeof(t_tab)*to->maxtabs);
}

extern void _CpriSetSoftMargin(t_tab_frame * tf, t_tab * margin, int newmargin,
                          float scaling)
{
   t_tab *m, tmp=*margin;
   int i;

   m = &tmp;
   m->mm = newmargin;
   if (m->mm < 0)
      m->mm = 0;
   m->scaled = (int)(scaling * m->mm + 0.5);
   m->block =(int)( m->mm / 2.54 + 0.5);
   for (i = 0; i < tf->n; i++) {
      tf->tabs[i].mm += m->mm - margin->mm;
      tf->tabs[i].scaled += m->scaled - margin->scaled;
      tf->tabs[i].block += m->block - margin->block;
   }
   *margin = *m;
}

extern char _CpriToDOS(char c, int symbolset)
{
   static char win2symb[2][256];
   static int first = 1;
   int j, i;

   if (first) {
      first = 0;
      for (j = 0; j < 2; j++)
         for (i = 0; i < 256; i++)
            win2symb[j][i] = (char)i;
#ifdef _WIN32
#ifndef __GNUC__
#pragma warning(push)
#pragma warning(disable : 4244; disable : 4305)
#endif
#endif
      win2symb[1][192] = 161;
      win2symb[1][194] = 162;
      win2symb[1][200] = 163;
      win2symb[1][202] = 164;
      win2symb[1][203] = 165;
      win2symb[1][206] = 166;
      win2symb[1][207] = 167;
      win2symb[1][217] = 173;
      win2symb[1][219] = 174;
      win2symb[1][221] = 177;
      win2symb[1][253] = 178;
      win2symb[1][226] = 192;
      win2symb[1][234] = 193;
      win2symb[1][244] = 194;
      win2symb[1][251] = 195;
      win2symb[1][225] = 196;
      win2symb[1][233] = 197;
      win2symb[1][243] = 198;
      win2symb[1][250] = 199;

      win2symb[1][224] = 200;
      win2symb[1][232] = 201;
      win2symb[1][242] = 202;
      win2symb[1][249] = 203;
      win2symb[1][228] = 204;
      win2symb[1][235] = 205;
      win2symb[1][246] = 206;
      win2symb[1][252] = 207;
      win2symb[1][197] = 208;
      win2symb[1][238] = 209;
      win2symb[1][216] = 210;
      win2symb[1][198] = 211;
      win2symb[1][229] = 212;
      win2symb[1][237] = 213;
      win2symb[1][248] = 214;
      win2symb[1][230] = 215;
      win2symb[1][196] = 216;
      win2symb[1][236] = 217;
      win2symb[1][214] = 218;
      win2symb[1][220] = 219;
      win2symb[1][201] = 220;
      win2symb[1][239] = 221;
      win2symb[1][223] = 222;
      win2symb[1][212] = 223;

      win2symb[1][193] = 224;
      win2symb[1][195] = 225;
      win2symb[1][227] = 226;
      win2symb[1][208] = 227;
      win2symb[1][240] = 228;
      win2symb[1][205] = 229;
      win2symb[1][204] = 230;
      win2symb[1][211] = 231;
      win2symb[1][210] = 232;
      win2symb[1][213] = 233;
      win2symb[1][245] = 234;
      win2symb[1][218] = 237;
      win2symb[1][255] = 239;
      win2symb[1][181] = 243;
      win2symb[1][190] = 245;
      win2symb[1][188] = 247;
      win2symb[1][189] = 248;
      win2symb[1][187] = 253;
      win2symb[0][161] = 173;
      win2symb[0][162] = 155;
      win2symb[0][163] = 156;
      win2symb[0][165] = 157;
      win2symb[0][170] = 166;
      win2symb[0][171] = 174;
      win2symb[0][176] = 248;
      win2symb[0][177] = 127;
      win2symb[0][178] = 253;
      win2symb[0][181] = 230;
      win2symb[0][186] = 167;
      win2symb[0][187] = 175;
      win2symb[0][188] = 172;
      win2symb[0][189] = 171;
      win2symb[0][191] = 168;
      win2symb[0][192] = 65;
      win2symb[0][193] = 65;
      win2symb[0][194] = 65;
      win2symb[0][195] = 65;
      win2symb[0][196] = 142;
      win2symb[0][197] = 143;
      win2symb[0][198] = 146;
      win2symb[0][199] = 128;
      win2symb[0][200] = 69;
      win2symb[0][201] = 149;
      win2symb[0][202] = 69;
      win2symb[0][203] = 69;
      win2symb[0][204] = 73;
      win2symb[0][205] = 73;
      win2symb[0][206] = 73;
      win2symb[0][207] = 73;
      win2symb[0][208] = 68;
      win2symb[0][209] = 165;
      win2symb[0][210] = 79;
      win2symb[0][211] = 79;
      win2symb[0][212] = 79;
      win2symb[0][213] = 79;
      win2symb[0][214] = 153;

      win2symb[0][216] = 153;
      win2symb[0][217] = 85;
      win2symb[0][218] = 85;
      win2symb[0][219] = 85;
      win2symb[0][220] = 154;
      win2symb[0][221] = 89;

      win2symb[0][223] = 225;
      win2symb[0][224] = 133;
      win2symb[0][225] = 160;
      win2symb[0][226] = 131;
      win2symb[0][227] = 97;
      win2symb[0][228] = 132;
      win2symb[0][229] = 134;
      win2symb[0][230] = 145;
      win2symb[0][231] = 135;
      win2symb[0][232] = 138;
      win2symb[0][233] = 130;
      win2symb[0][234] = 136;
      win2symb[0][235] = 137;
      win2symb[0][236] = 141;
      win2symb[0][237] = 161;
      win2symb[0][238] = 140;
      win2symb[0][239] = 139;

      win2symb[0][241] = 164;
      win2symb[0][242] = 149;
      win2symb[0][243] = 162;
      win2symb[0][244] = 147;
      win2symb[0][245] = 111;
      win2symb[0][246] = 148;
      win2symb[0][247] = 246;
      win2symb[0][248] = 148;
      win2symb[0][249] = 151;
      win2symb[0][250] = 163;
      win2symb[0][251] = 150;
      win2symb[0][252] = 129;
      win2symb[0][253] = 121;

      win2symb[0][255] = 152;
#ifdef _WIN32
#ifndef __GNUC__
#pragma warning(pop)
#endif
#endif

   }
   if ((unsigned char)c > 127)
      symbolset += 2;
   if (c < 0)
      return c;
   return win2symb[(symbolset - 1) & 1][c & 0xff];
}

extern const char *const*CpriGetTypefaceNames(void)
{
   return typeface_names;
}
