/* Module DRV_HP.C
   Contains the print-driver for HP-printers */
#ifdef DJGPP
#include <string.h>
#include <allegro.h>

#include <cgui.h>
#include <cgui/mem.h>
#include "labldata.h"

#include "cpri.h"
#include "spooler.h"
#include "pridrive.h"

#include "drv_hp.h"
#include "drv_txt.h"
#include "drv_common.h"
#include "width.h"
#include "dialpri.h"
#include "cpri.ht"

#define IndexSize(size) ( (size)*2/3 )
#define HP_PATTERN_WIDTH_LOWRES  "\33*c%dA"
#define HP_PATTERN_HEIGHT_LOWRES "\33*c%dB"
#define PRINT_BLACK_PATTERN "\33*c0P" /* Prints the currently specified
                     pattern area at current position using black color */
#define HP_BARWIDTH "\33*c%dH"
#define HP_BARHEIGHT "\33*c%dV"
#define HP_SCALING         300
#define HP_FINE_SCALING    720
#define HP_FIXCHAR          30
#define HP_MM               12
#define HP_FIXHEIGHT        50
#define ID_TOPLINE        1001
#define ID_ROWGRID        1002
#define A4_PAGE_HEIGHT    3508
#define A4_PAGE_WIDTH     2480
#define HP_UNDERLINE_ON "\33&d3D"
#define HP_UNDERLINE_OFF "\33&d@"
#define ROW_HEIGHT(dd) ((int)((dd->size/2.834645)*dd->scaling+0.5))
#define HP_MARGINS "\33&l26A\0339\33&l0L\33&l0E\33&l0F"
#define HP_LINES_PER_INCH "\33&l6D"
#define HP_SYMBOLSET "\33(s10H\33(9U"
#define HP_MACROID "\33&f%dY"
#define HP_ABSMOVEX "\33*p%dX"
#define HP_ABSMOVEY "\33*p%dY"
#define HP_RELMOVEX "\33*p%+dX"
#define HP_RELMOVEY "\33*p%+dY"
#define HP_MOVERIGHT "\33&a%+dH"
#define HP_MOVEDOWN "\33&a%+dV"
#define HP_TYPEFACES "\33(s%dP\33(s%dT\33(s%dB\33(s%dS"
#define HP_DELETEMACROS "\33&f6X"
#define HP_SIZE "\33(s%dV"
#define HP_PORTRAIT "\33&l0O"
#define HP_LANDSCAPE "\33&l1O"
#define HP_SEMIGRAPH_ON "\xE"
#define HP_SEMIGRAPH_OFF "\xF"
#define HP_MANUALFEED "\33&l2H"
#define HP_CASSETTEFEED "\33&l1H"
#define HP_PUSHPOS "\33&f0S"
#define HP_POPPOS "\33&f1S"
#define HP_STARTMACRO "\33&f0X"
#define HP_STOPMACRO "\33&f1X"
#define HP_RUNMACRO "\33&f3X"
#define BASELINE_TO_CELLBOTTOM(ts) ((ts)->rowheight/2 - 26*(ts)->rowheight/(ts)->hightpercent)
#define CELLTOP_TO_BASELINE(ts) (26*(ts)->rowheight/(ts)->hightpercent + (ts)->rowheight/2)
#define FONT_HEIGHT(dd) (((dd)->tstate->size / 2.834645) * (dd)->scaling + 0.5)
#define FIX_SPACING  0
#define PROP_SPACING 1
#define EDGE_DIST 20       /* The start point for embedded text (within bar) */
#define REPEAT_HEADER 0    /* A type of macro (n.u.) */

/* These values are considered as table related, i.e. they have to be set
   before starting to play a macro in order to give it the same environment
   to tun in as when recorded */
typedef struct t_tblstate {
   t_tab_frame tf;     /* A list of the current tabs */
   int rowheight;      /* The height of each row (line feed rel-ypos) */
   int hightpercent;   /* The row-heigth expressed in % of current fontsize */
   int fixrowheight;   /* If non-zero the height of the rows will take this value */
   int linepospercent; /* The displacement as received from the user (%) */
   int text2grid;      /* The displacement of the text in the cell */
   int gridon;         /* Tells if cells is to be outlined */
   int size;           /* Font size */
   int style;          /* Font style */
   int stroke;         /* Stroke weight of font */
   int face;           /* Font typeface */
   int ul;             /* Indicates if the underline option is on */
   int newtable;       /* Indicates that a new table has been specified */
} t_tblstate;

typedef struct t_macro {
   int id;                       /* Macro id. (n.u) */
   int type;                     /* Type of macor (n.u) */
   t_tblstate tstate;            /* The table state at the start point of
                                    recording. Will be set when playing starts */
   t_tblstate tmptstate;         /* The table state when starting to play the
                                    macro. Will be restored when playing stops */
   char *start;                  /* The start and end positions in the input */
   char *orig;                   /* The input position when playing starts.
                                    Will be restored after playing finished
                                    (since playing just redirects input
                                    position) */
} t_macro;

typedef struct t_driver {
   struct t_printerq *q;
   char *outptr;                /* A pointer to next character in outmembuf
                                   to sent to the printer */
   char *outend;                /* A pointer to the end of the output string.
                                   Used during the process of boulding the
                                   output string */
   char *outmembuf;             /* A string containg a certain amount of
                                   text and PCL commands prepared for printing. */
   char *instart;               /* A pointer to beginning of memory buffered
                                   input data */
   t_macro *macros;             /* An array of 'nmacros' number of macros */
   int nmacros;                 /* Number of current macros */
   t_macro *current_record_macro;/* The macro currently recorded */
   t_macro *current_play_macro; /* The macro currently played */
   int playing;                 /* Indicates if playing of macro is in progress */
   int recording;               /* Indicates if recording of macro is in progress */
   int macroids;                /* A server for generating unique names */
   t_tblstate *tstate;          /* Stuff related to the state of the table */
   t_tblstate tstatemem;        /* Pointer to above; may be redirected */
   int tabletop;                /* The top of the curtrent table */
   int tablebottom;             /* The bottom of the curtrent table */
   double scaling;              /* Scaling factor (currently values will be
                                   scaled from text-box or mm to suitabble
                                   values for the printer) */
   int ready;
   double fine_scaling;         /* Alternative scaling used in when higher
                                   acuracy of measurment is needed */
   int ypos;                    /* current printing y-position */
   int gridinited;              /* Is used to detect if the upper part of the
                                   grid has to be drawn */
   int indexwidth;
   int macroplaying;
   int lframesp;                /* space between the left of the text and the
                                   frame */
   int nocopies;                /* number of copies of each printout */
   int copies;                  /* flag indicationg that printing of copies
                                   is in progress */
   int paper_height;            /* Current page hight. Depending on orienta-
                                   tion and paper size (currently only A4) */
   int xt_fonts;                /* extended fonts in the current printer
                                   brand */
   int dos;                     /* flag indicating if character codes are to
                                   be converted to the PC-8 character set */
   char *nspace;                /* different barcodes for different printers */
   char *wspace;
   char *wbar;
   int bc_height;               /* The current height of the bars in bar-codes */
   int wide_width;              /* The current width of the wide bars in bar-codes */
   int narrow_width;            /* The current width of the narrow bars in bar-codes */
   int printing_embedded;
   char *embedded_string;
   char *postponed_superscript;
} t_driver;

static void ProcessComponent(t_driver *dd);

static char *print_font(char *p, int face, int stroke, int style)
{
   /* HP # codes for style {upright, italic} */
   static int styles[][2] = {
          {0, 0}, {0, 1}, {0, 1}, {0, 1}, {4, 5}, {0, 1}, {0, 1}, {0, 1},
          {0, 1}, {0, 0}, {2, 2}
   };

   /* HP # codes for stroke {thin, bold} */
   static int strokes[][2] = {
          {1, 4}, {0, 2}, {0, 3}, {0, 3}, {0, 1}, {0, 3}, {0, 1}, {0, 2},
          {0, 0}, {0, 0}, {2, 2}
   };
   int facecode, spacing;

   spacing = PROP_SPACING;
   switch (face) {
   case ALBERTUS:
      facecode = 0;
      style = 0;
      break;
   case ANTIQUE:
      facecode = 72;
      if (stroke && style)
         style = 0;
      break;
   case TIMES:
      facecode = 5;
      break;
   case UNIVERS:
      facecode = 52;
      break;
   case CONDENSED:
      facecode = 52;
      break;
   case FONT5:
      facecode = 17;
      break;
   case COURIER:
      facecode = 3;
      spacing = FIX_SPACING;
      break;
   case GOTHIC:
      facecode = 6;
      spacing = FIX_SPACING;
      break;
   case SCRIPT:
      facecode = 20;
      style = 1;
      stroke = 0;
      break;
   case FONT10:
      facecode = 0;
      style = 1;
      stroke = 0;
      break;
   case FONT11:
      facecode = 0;
      style = 0;
      stroke = 1;
   default:
      facecode = 0;
      break;
   }
   p += sprintf(p, HP_TYPEFACES, spacing, facecode, strokes[face][stroke],
                   styles[face][style]);
   return p;
}

static char *DrawRectAtCurrentPos(char *p, double fwidth, double fheight)
{
   int width, height;

   width = fwidth;
   height = fheight;
   /* Top line */
   p += sprintf(p, HP_PATTERN_WIDTH_LOWRES, width);
   p += sprintf(p, HP_PATTERN_HEIGHT_LOWRES, 2);
   p += sprintf(p, PRINT_BLACK_PATTERN);
   /* Bottom line */
   p += sprintf(p, HP_RELMOVEY, height);
   p += sprintf(p, PRINT_BLACK_PATTERN);
   p += sprintf(p, HP_RELMOVEY, -height);
   /* Left line */
   p += sprintf(p, HP_PATTERN_WIDTH_LOWRES, 2);
   p += sprintf(p, HP_PATTERN_HEIGHT_LOWRES, height);
   p += sprintf(p, PRINT_BLACK_PATTERN);
   /* Right line */
   p += sprintf(p, HP_RELMOVEX, width);
   p += sprintf(p, PRINT_BLACK_PATTERN);
   return p;
}

static char *PrintBarCodeAtCurPos(t_driver *dd, char *p, char *s)
{
   /* Printing barcodes of type INTERLEAVED 2/5 with HP-Lasejet-III commands.
      Interleaved 2/5 is capable to print an arbitrary long  sequence of an
      even number of digits (0..9). The digits are printed as pairs, the
      first one in a pair constitutes the bars, and the second one
      constitutes the spaces between the bars in the former one. The code is
      self-correcting with "1.4 bit" (i.e. in all cases there is at least
      one bit used for self correcting and in 40% of the cases there are 2).
      The code has different start- and stopsequences.

      The HP-implementation uses "black pattern" for printing a bar. Black
      pattern simply means "filled rectangle" in the PCL language. It works
      like this: Specify width and height of the rectangle and then send the
      "print black pattern" command, thats all (the print position is not
      advanced). We are supposed to print the code at current position, the
      bar code beeing above and to the right. The "print black pattern"
      command prints the rectangle below the current position, so we
      need to start a bit higher (i.e. the height of the bar code) and
      then restore this at the end.
      The spaces are implementedd just as movement. The bar code printing
      uses the higher resolution of the printer since it is quite important
      with correct proportions. The actual printed area of a filled
      rectangle is estimated (just some random tests) to get about 3 pixels
      wider around the entire rectangle, probably because the toner flow over
      the edges a bit. Therefore it is an offset of 6 for the movement. Maybe
      some other value is more acurate using non standard paper. */

   /* Table defining the pattern of one digit
      0 - narrow bar or space
      1 - wide bar or space */
   static int bc[10][5] = {
      {0, 0, 1, 1, 0},
      {1, 0, 0, 0, 1},
      {0, 1, 0, 0, 1},
      {1, 1, 0, 0, 0},
      {0, 0, 1, 0, 1},
      {1, 0, 1, 0, 0},
      {0, 1, 1, 0, 0},
      {0, 0, 0, 1, 1},
      {1, 0, 0, 1, 0},
      {0, 1, 0, 1, 0}
   };
   int i, j, n, width;

   n = strlen(s);
   if (n & 1) { /* Reject if not even number of digits */
      p += sprintf(p, "%s is not an even number of digits", s);
      return p;
   }

   /* Set start value of the width to narrow */
   width = dd->narrow_width;

   /* Move the current position up, since the specified y-pos referes to
   the bottom of the bar codes */
   p += sprintf(p, HP_MOVEDOWN, -dd->bc_height);

   /* Set the bar-code height (i.e. specify the current pattern height) */
   p += sprintf(p, HP_BARHEIGHT, dd->bc_height);

   /* The startsequence of bars (narrow bar, narrow space, narrow bar, narrow space): */
   p += sprintf(p, HP_BARWIDTH, width);
   p += sprintf(p, PRINT_BLACK_PATTERN);
   p += sprintf(p, HP_MOVERIGHT, width*2 + 6);
   p += sprintf(p, PRINT_BLACK_PATTERN);
   p += sprintf(p, HP_MOVERIGHT, width*2 + 6);

   /* Print two digits at a time */
   for (i = 0; i < n; i += 2) {
      for (j = 0; j < 5; j++) {
         if (bc[s[i] - '0'][j]) { /* wide bar */
            if (width != dd->wide_width) {
               width = dd->wide_width;
               p += sprintf(p, HP_BARWIDTH, width);
            }
         } else {                  /* narrow bar */
            if (width != dd->narrow_width) {
               width = dd->narrow_width;
               p += sprintf(p, HP_BARWIDTH, width);
            }
         }
         p += sprintf(p, PRINT_BLACK_PATTERN);
         /* Add 6 because the bar will in practice be appr. 6 dots wider */
         if (bc[s[i + 1] - '0'][j]) /* wide space */
            p += sprintf(p, HP_MOVERIGHT, width + dd->wide_width + 6);
         else                   /* narrow space */
            p += sprintf(p, HP_MOVERIGHT, width + dd->narrow_width + 6);
      }
   }

   /* the stopsequence of bars: */
   p += sprintf(p, HP_BARWIDTH, dd->wide_width);
   p += sprintf(p, PRINT_BLACK_PATTERN);
   p += sprintf(p, HP_MOVERIGHT, dd->wide_width + dd->narrow_width + 6);
   p += sprintf(p, HP_BARWIDTH, dd->narrow_width);
   p += sprintf(p, PRINT_BLACK_PATTERN);

   /* Restore the y-pos as were before the call and advanced x-pos to end */
   p += sprintf(p, HP_MOVEDOWN, dd->bc_height);
   p += sprintf(p, HP_MOVERIGHT, dd->narrow_width);
   return p;
}

static char *SetTypface(t_driver *dd, char *p)
{
   static int x2basic[] = { UNIVERS, UNIVERS, TIMES, UNIVERS, UNIVERS,
         TIMES, COURIER, COURIER, UNIVERS, TIMES, TIMES };

   _CpriGetIntParam(dd->q, &dd->tstate->face);
   _CpriGetIntParam(dd->q, &dd->tstate->stroke);
   dd->tstate->stroke &= 1;
   _CpriGetIntParam(dd->q, &dd->tstate->style);
   dd->tstate->style &= 1;
   if ((unsigned) dd->tstate->face >= 11)
      dd->tstate->face = 2;
   if (!dd->xt_fonts) /* basic fonts only consist of TIMES,UNIVERS and COURIER */
      dd->tstate->face = x2basic[dd->tstate->face];
   p = print_font(p, dd->tstate->face, dd->tstate->stroke, dd->tstate->style);
   return p;
}

static char *RightAlignSuperscriptToCurrentPosition(t_driver *dd, char *p, const char *s)
{
   int indexsize, width, vertical_move;
   
   indexsize = IndexSize(dd->tstate->size);
   vertical_move = indexsize * 2;
   width = string_width(s, dd->tstate->face, dd->tstate->stroke, dd->tstate->style, indexsize, dd->xt_fonts, HP_SCALING);
   p += sprintf(p, HP_PUSHPOS);
   p += sprintf(p, HP_RELMOVEY, -vertical_move);
   p += sprintf(p, HP_RELMOVEX, -width);
   p += sprintf(p, HP_SIZE, indexsize);
   p += sprintf(p, s);
   p += sprintf(p, HP_SIZE, dd->tstate->size);
   p += sprintf(p, HP_POPPOS);
   return p;
}

static char *MoveToRightCellEdge(t_tab_frame *tf, char *p)
{
   int nextpos;
   
   nextpos = tf->tabs[tf->curtab].scaled;
   nextpos -= 25;
   p += sprintf(p, HP_ABSMOVEX, nextpos);
   return p;
}

static void GetPostponedSuperscript(t_driver *dd, char *tmp)
{
   char *curpos;
   curpos = _CpriGetInputPos(dd->q);
   _CpriSetInputPos(dd->q, dd->postponed_superscript);
   _CpriGetStringParam(dd->q, tmp);
   dd->postponed_superscript = NULL;
   _CpriSetInputPos(dd->q, curpos);
}

static char *AlignToRightCellEdge(t_driver *dd, char *p, unsigned char *s, int w)
{
   char tmp[100];

   p = MoveToRightCellEdge(&dd->tstate->tf, p);
   p += sprintf(p, HP_RELMOVEX, -w);
   if (dd->postponed_superscript) {
      GetPostponedSuperscript(dd, tmp);
      p = RightAlignSuperscriptToCurrentPosition(dd, p, tmp);
   }
   p += sprintf(p, s);
   return p;
}

static char *AlignHalfToRightCellEdge(t_driver *dd, char *p, unsigned char *s, int w)
{
   int len;

   len = strlen(s);
   if (s[len - 1] != (unsigned char) HALF_CHAR || len == 1)
      w += string_width(HALF_STRING, dd->tstate->face, dd->tstate->stroke, dd->tstate->style, dd->tstate->size, dd->xt_fonts, HP_SCALING);
   p = AlignToRightCellEdge(dd, p, s, w);
   return p;
}

static char *AlignDeciToRightCellEdge(t_driver *dd, char *p, unsigned char *s, int w)
{
   int len;

   len = strlen(s);
   if (len > 1) {
      if (s[len - 2] == '.');
      else if (s[len - 1] == '.')
         w += string_width("0", dd->tstate->face, dd->tstate->stroke, dd->tstate->style, dd->tstate->size, dd->xt_fonts, HP_SCALING);
      else
         w += string_width(".0", dd->tstate->face, dd->tstate->stroke, dd->tstate->style, dd->tstate->size, dd->xt_fonts, HP_SCALING);
   } else if (len == 1) {
      if (*s == '.')
         w += string_width("0", dd->tstate->face, dd->tstate->stroke, dd->tstate->style, dd->tstate->size, dd->xt_fonts, HP_SCALING);
      else
         w += string_width(".0", dd->tstate->face, dd->tstate->stroke, dd->tstate->style, dd->tstate->size, dd->xt_fonts, HP_SCALING);
   }
   p = AlignToRightCellEdge(dd, p, s, w);
   return p;
}

static char *Center(t_driver *dd, char *p, unsigned char *s, int w)
{
   char tmp[100];
   int centerpoint;
   t_tab_frame *tf = &dd->tstate->tf;

   if (tf->curtab < 1 || tf->curtab >= tf->n)
      return p;
   centerpoint = (tf->tabs[tf->curtab].scaled + tf->tabs[tf->prevtab].scaled) / 2;
   p += sprintf(p, HP_ABSMOVEX, centerpoint - w/2);
   if (dd->postponed_superscript) {
      GetPostponedSuperscript(dd, tmp);
      p = RightAlignSuperscriptToCurrentPosition(dd, p, tmp);
   }
   p += sprintf(p, s);
   return p;
}

static char *SuperscriptAtCurrentPos(t_driver *dd, char *p, unsigned char *s, int fontsize)
{
   int vertical_move, indexsize;

   if (dd->tstate->ul)
      p += sprintf(p, HP_UNDERLINE_OFF);
   indexsize = IndexSize(fontsize);
   vertical_move = indexsize * 2;
   p += sprintf(p, HP_RELMOVEY, -vertical_move);
   p += sprintf(p, HP_SIZE, indexsize);
   p += sprintf(p, s);
   p += sprintf(p, HP_SIZE, fontsize);
   p += sprintf(p, HP_RELMOVEY, vertical_move);
   if (dd->tstate->ul)
      p += sprintf(p, HP_UNDERLINE_ON);
   return p;
}

static char *SuperscriptRightAligned(t_driver *dd, char *p, unsigned char *s)
{
   p = MoveToRightCellEdge(&dd->tstate->tf, p);
   p = RightAlignSuperscriptToCurrentPosition(dd, p, s);
   return p;
}

static char *MakeDiagramBar(t_driver *dd, char *p, int width, int h,
                    int pattern, int grey, int embedded)
{
   t_tblstate *ts;
   int strw, c;
   char *s, tmpbuf[1000];

   ts = dd->tstate;
   
   dd->printing_embedded = 1;
   if (!embedded)
      p += sprintf(p, HP_PUSHPOS);
   
   strw = 0;
   do {
      dd->outend = p;
      ProcessComponent(dd);
      p = dd->outend;
      /* Calculate the total text length */
      if (dd->printing_embedded) {
         for (s = dd->embedded_string; *s && *s!=27; s++)
            ;
         c = *s;
         *s = 0;
         strw += string_width(dd->embedded_string, dd->tstate->face,
                 dd->tstate->stroke, dd->tstate->style, dd->tstate->size, dd->xt_fonts, HP_SCALING);
         *s = c;
      }
   } while (dd->printing_embedded);
   
   width = (width * 3000) / 254;
   h = (h * 3000) / 254;
   if (h < FONT_HEIGHT(dd))
      h = FONT_HEIGHT(dd);
   if (ts->rowheight < h)
      ts->rowheight = h;

   if (embedded) {
      strcpy(tmpbuf, dd->outmembuf);
      p = dd->outend = dd->outmembuf;
      p += sprintf(p, HP_PUSHPOS);
      p += sprintf(p, HP_RELMOVEY, (ts->rowheight - h) / 2);
   } else {
      p += sprintf(p, HP_POPPOS);
      p += sprintf(p, HP_PUSHPOS);
      p += sprintf(p, HP_RELMOVEY, (int)(3*FONT_HEIGHT(dd)/4));
   }
   p += sprintf(p, HP_PUSHPOS);
   /* Print the bar */
   p += sprintf(p, HP_PATTERN_WIDTH_LOWRES, width);
   p += sprintf(p, HP_PATTERN_HEIGHT_LOWRES, h);
   if (pattern == BAR_PATTERN_GRAY)
      p += sprintf(p, "\33*c%dG\33*c2P", grey);
   else
      p += sprintf(p, "\33*c%dG\33*c3P", pattern);

   if (embedded) {
      p += sprintf(p, HP_RELMOVEY, (int)((h - FONT_HEIGHT(dd)) / 2));
      if (strw + 22 > width  && width + strw + 12 < ts->tf.tabs[ts->tf.curtab+1].scaled) {
         p += sprintf(p, HP_RELMOVEX, width + 12);
      } else {
         /* Draw white rectangle */
         p += sprintf(p, HP_RELMOVEX, 12);
         p += sprintf(p, HP_PATTERN_WIDTH_LOWRES, strw + 10);
         p += sprintf(p, HP_PATTERN_HEIGHT_LOWRES, (int)FONT_HEIGHT(dd));
         p += sprintf(p, "\33*c1P");
         p += sprintf(p, HP_RELMOVEX, 5);
      }
      /* The buffered text */
      p += sprintf(p, HP_RELMOVEY, (int)(3*FONT_HEIGHT(dd)/4));
      p += sprintf(p, tmpbuf);
   }
   p += sprintf(p, HP_POPPOS);
   p = DrawRectAtCurrentPos(p, width, h);
   p += sprintf(p, HP_POPPOS);
   return p;
}

/* Assumes that x position is at left edge of table and ypos is on
   the base line of the text within that row. Will not affect the print
   position.
   - Store curpos
   - Move ypos from baseline to top of row.
   - Draws a horizontal line of tablewidth length
   - Restore curpos */
static char *DrawUpperGridSegment(t_driver *dd, char *p)
{
   int width=0;
   t_tblstate *ts;
   t_tab_frame *tf;

   ts = dd->tstate;
   tf = &ts->tf;
   p += sprintf(p, HP_PUSHPOS);
   if (ts->hightpercent)
      p += sprintf(p, HP_RELMOVEY, -CELLTOP_TO_BASELINE(ts));
   if (tf->n > 1)
      width = tf->tabs[tf->n-1].scaled - tf->tabs[0].scaled;
   p += sprintf(p, HP_PATTERN_WIDTH_LOWRES, width);
   p += sprintf(p, HP_PATTERN_HEIGHT_LOWRES, 3);
   p += sprintf(p, PRINT_BLACK_PATTERN); /* Draw horizontal line */
   p += sprintf(p, HP_POPPOS);
   return p;
}

/* Assumes that x position is at left edge of table and ypos is on
   the base line of the text within that row. Will not affect the print
   position.
   - Store curpos
   - Move ypos from baseline to bottom of row.
   - Draws a horizontal line of tablewidth length
   - Move up to top of row.
   - For each column: draw a vertical line of rowheight height
   - Restore curpos */
static char *DrawLowerGridSegment(t_driver *dd, char *p)
{
   int width=0, i;
   t_tblstate *ts;
   t_tab_frame *tf;

   ts = dd->tstate;
   tf = &ts->tf;
   p += sprintf(p, HP_PUSHPOS);
   if (ts->hightpercent)
      p += sprintf(p, HP_RELMOVEY, BASELINE_TO_CELLBOTTOM(ts));
   if (tf->n > 1)
      width = tf->tabs[tf->n-1].scaled - tf->tabs[0].scaled;
   p += sprintf(p, HP_PATTERN_WIDTH_LOWRES, width);
   p += sprintf(p, HP_PATTERN_HEIGHT_LOWRES, 3);
   p += sprintf(p, PRINT_BLACK_PATTERN); /* Draw horizontal line */
   p += sprintf(p, HP_PATTERN_WIDTH_LOWRES, 3);
   p += sprintf(p, HP_PATTERN_HEIGHT_LOWRES , ts->rowheight);
   p += sprintf(p, HP_RELMOVEY, -ts->rowheight);
   for (i=0; i<tf->n; i++) {
      if (!tf->tabs[i].inactive) {
         p += sprintf(p, HP_ABSMOVEX, tf->tabs[i].scaled);
         p += sprintf(p, PRINT_BLACK_PATTERN); /* Draw horizontal line */
      }
   }
   p += sprintf(p, HP_POPPOS);
   return p;
}

static void ClearTabs(t_driver *dd)
{
   dd->tstate->tf.n = 0;
}

static void DestroyMacro(t_macro *m)
{
   _CpriDestroyTabFrame(&m->tstate.tf);
}

static void CopyTableState(t_tblstate *from, t_tblstate *to)
{
   *to = *from;
   _CpriCopyTabFrame(&from->tf, &to->tf);
}

static char *SetDifferingStates(char *p, t_tblstate *cur, t_tblstate *ts)
{
   if (ts->style != cur->style || ts->stroke != cur->stroke || ts->face != cur->face)
      p = print_font(p, ts->face, ts->stroke, ts->style);
   if (ts->size != cur->size)
      p += sprintf(p, HP_SIZE, ts->size);
   return p;
}

static void CreateNewMacro(t_driver *dd)
{
   t_macro *m;

   dd->nmacros++;
   dd->macros = ResizeMem(t_macro, dd->macros, dd->nmacros);
   m = dd->macros + dd->nmacros - 1;
   m->id = dd->macroids++;
   m->type = REPEAT_HEADER;
   m->start = _CpriGetInputPos(dd->q);
   dd->current_record_macro = m;
}

static char *StartPlayMacro(t_driver *dd, char *p)
{
   t_macro *m;

   m = dd->current_play_macro;
   p = SetDifferingStates(p, dd->tstate, &m->tstate);
   CopyTableState(&m->tstate, &m->tmptstate);
   dd->tstate = &m->tmptstate;
   m->orig = _CpriGetInputPos(dd->q);
   _CpriSetInputPos(dd->q, m->start);
   dd->playing = 1;
   return p;
}

static char *MoveToNextCell(t_driver *dd, char *p);
static char *StopPlayMacro(t_driver *dd, char *p)
{
   t_macro *m;
   t_tab_frame *tf;
   
   m = dd->current_play_macro;
   dd->tstate = &dd->tstatemem;
   tf = &dd->tstate->tf;
   p = SetDifferingStates(p, &m->tmptstate, dd->tstate);
   _CpriDestroyTabFrame(&m->tmptstate.tf);
   /* This compensates for the effect of starting the recording after the
   first MOVE_TO_NEXT_COLUMN */
   tf->curtab = 0;
   if (dd->tstate->gridon && !m->tmptstate.gridon)
      dd->gridinited = 0;
   p = MoveToNextCell(dd, p);
   _CpriSetInputPos(dd->q, m->orig);
   dd->playing = 0;
   return p;
}

static char *PlayNextMacro(t_driver *dd, char *p)
{
   t_macro *m;

   m = dd->current_play_macro;
   if (m == dd->macros + dd->nmacros - 1) {
      dd->current_play_macro = NULL;
      return p; /* No more macros to play */
   }
   m++;
   p = StartPlayMacro(dd, p);
   return p;
}

static void StartRecordMacro(t_driver *dd)
{
   dd->recording = 1;
   CreateNewMacro(dd);
   CopyTableState(dd->tstate, &dd->current_record_macro->tstate);
}

static char *StopRecordMacro(t_driver *dd, char *p)
{
   if (dd->recording) {
      dd->recording = 0;
   } else if (dd->playing) {
      StopPlayMacro(dd, p);
      PlayNextMacro(dd, p);
   }
   return p;
}

static char *NewPage(t_driver *dd, char *p)
{
   dd->gridinited = 0; /* Forces the top line to be printed */
   p += sprintf(p, "\14\15");
   dd->ypos = dd->tabletop;
   p += sprintf(p, HP_ABSMOVEY, dd->ypos);
   if (dd->nmacros) {
      dd->current_play_macro = dd->macros + 0; /* launch first macro */
      p = StartPlayMacro(dd, p);
   }
   p += sprintf(p, HP_ABSMOVEX, dd->tstate->tf.tabs[0].scaled);
   return p;
}

static char *NewLine(t_driver *dd, char *p)
{
   dd->tstate->tf.curtab = 0;
   if (dd->tstate->ul)
      p += sprintf(p, HP_UNDERLINE_OFF);
   p += sprintf(p, HP_ABSMOVEX, dd->tstate->tf.tabs[0].scaled);
   if (dd->tstate->gridon)
      p = DrawLowerGridSegment(dd, p);
   if (dd->tstate->ul)
      p += sprintf(p, HP_UNDERLINE_ON);
   if (dd->recording)
      p = StopRecordMacro(dd, p);
   return p;
}

static char *MakeStrikeText(t_driver *dd, char *p, unsigned char *s)
{
   int w, h;

   w = string_width(s, dd->tstate->face, dd->tstate->stroke, dd->tstate->style, dd->tstate->size, dd->xt_fonts, HP_SCALING);
   h = FONT_HEIGHT(dd) / 4;
   p += sprintf(p, HP_RELMOVEY, -h);
   p += sprintf(p, HP_PATTERN_WIDTH_LOWRES, w);
   p += sprintf(p, HP_PATTERN_HEIGHT_LOWRES, 3);
   p += sprintf(p, PRINT_BLACK_PATTERN);
   p += sprintf(p, HP_RELMOVEY, h);
   p += sprintf(p, s);
   return p;
}

static int SetLineHeight(t_driver *dd)
{
   int ymove;

   if (dd->tstate->fixrowheight)
      return dd->tstate->rowheight = dd->tstate->fixrowheight;
   ymove = BASELINE_TO_CELLBOTTOM(dd->tstate);
   dd->tstate->rowheight = (dd->tstate->size / 2.834645) * (dd->tstate->hightpercent / 100.0) * dd->scaling + 0.5;
   dd->tstate->text2grid = dd->tstate->rowheight * dd->tstate->linepospercent / 100.0 + 0.5;
   ymove += CELLTOP_TO_BASELINE(dd->tstate);
   return ymove;
}

static void SetBarCodeHeight(t_driver *dd, int h)
{
   dd->bc_height = h * dd->fine_scaling;
}

static void SetBarCodeWidth(t_driver *dd, int w)
{
   dd->narrow_width = w * HP_FINE_SCALING/2540. + 0.5;
   switch (dd->narrow_width) {
   case 1:
      dd->wide_width = dd->narrow_width * 7;
      break;
   case 2:
   case 3:
      dd->wide_width = dd->narrow_width * 5;
      break;
   case 4:
   case 5:
      dd->wide_width = dd->narrow_width * 4;
      break;
   case 6:
   case 7:
      dd->wide_width = dd->narrow_width * 3;
      break;
   default:
      dd->wide_width = dd->narrow_width * 2;
      break;
   }
}

static void InitDriverData(t_driver *dd)
{
   dd->tstate = &dd->tstatemem;
   dd->scaling = HP_SCALING / 25.4;
   dd->fine_scaling = HP_FINE_SCALING / 25.4;
   dd->lframesp = 2 * HP_MM;
   dd->tstate->rowheight = 100;
   dd->tstate->size = 12;
   dd->tstate->face = COURIER;
   dd->tablebottom = A4_PAGE_HEIGHT - 200;
   dd->paper_height = A4_PAGE_HEIGHT;
   dd->tstate->linepospercent = 50;
   dd->tstate->hightpercent = 100;
   dd->tstate->fixrowheight = 0;
   dd->postponed_superscript = NULL;
   dd->nocopies = 1;
   SetBarCodeWidth(dd, 7);
   dd->bc_height = 8 * dd->fine_scaling; /* 8mm = default height */

   dd->outmembuf = GetMem(char, 1000);

   dd->outend = dd->outptr = dd->outmembuf;
   *dd->outend = 0;
   if (dd->xt_fonts) {
      dd->nspace = "\33*p-49X";
      dd->wspace = "\33*p-41X";
      dd->wbar = "\263\33*p-55X\263\33*p-55X\263";
   } else {
      dd->nspace = "\33*p-73X";
      dd->wspace = "\33*p-65X";
      dd->wbar = "\263\33*p-79X\263\33*p-79X\263";
   }
}

static struct t_driver *HPInit(struct t_printerq *q)
{
   t_driver *dd;

   dd = GetMem0(t_driver, 1);
   InitDriverData(dd);
   dd->q = q;
   return dd;
}

static struct t_driver *KyoceraInit(struct t_printerq *q)
{
   t_driver *dd;

   dd = HPInit(q);
   dd->xt_fonts = 1;
   return dd;
}

static struct t_driver *BrotherInit(struct t_printerq *q)
{
   return KyoceraInit(q);
}

static void HPClose(t_driver *dd)
{
   if (dd) {
      if (dd->outmembuf)
         Release(dd->outmembuf);
      _CpriDestroyTabFrame(&dd->tstate->tf);
      Release(dd);
   }
}

extern void HPTest(int qno)
{
   t_handle hnd;
   const char * const * names;
   int i, j, k, nr;

   /* char block[]={127,177}; */
   int sizes[] = { 8, 10, 12, 14, 18, 24, 36, 50, 100 };
   int height = 354;
   int width1 = 80;
   int width2 = 60;
   const char *const*ptxt;

   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      exit(-1);
   hnd = CpriOpenEntry(qno, "Print test", 0, "");
   if (hnd) {

      /* port info */
      CpriCommand(hnd, CPRI_PORTRAIT);
      CpriCommand(hnd, CPRI_TABLE, 20, 20, 10);
      CpriCommand(hnd, CPRI_LINE_HEIGHT, 120);
      CpriCommand(hnd, CPRI_FONT_TYPEFACE, TIMES, 0, 0);
      CpriCommand(hnd, CPRI_FONT_SIZE, 12);
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriText(hnd, ptxt[pt_print_test_on]);
      CpriText(hnd, ptxt[pt_selq + qno]);
      CpriCommand(hnd, CPRI_NEW_LINE);
//      CpriText(hnd, ptxt[pt_pribrands + brand]);
      CpriCommand(hnd, CPRI_NEW_LINE);

      /* table test */
      CpriCommand(hnd, CPRI_NEW_LINE);
      /* CpriText(hnd, "%c", block[q->dostext]); */
      CpriCommand(hnd, CPRI_REMOVE_COLUMNS);
      CpriCommand(hnd, CPRI_LINE_HEIGHT, height);
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCommand(hnd, CPRI_TEXT_TO_HLINE_SPACE, 50);
      CpriCommand(hnd, CPRI_CREATE_COLUMN, width1);
      CpriCommand(hnd, CPRI_CREATE_COLUMN, width2);
      CpriCommand(hnd, CPRI_START_GRID);
      CpriCommand(hnd, CPRI_MOVE_TO_NEXT_COLUMN);
      CpriText(hnd, ptxt[pt_text_in_frame_1]);
//      CpriText(hnd, " %d", width1);
      CpriCommand(hnd, CPRI_MOVE_TO_NEXT_COLUMN);
      CpriText(hnd, ptxt[pt_text_in_frame_2]);
//      CpriText(hnd, " %d", width2);
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCommand(hnd, CPRI_MOVE_TO_NEXT_COLUMN);
      CpriText(hnd, ptxt[pt_text_frame_height]);
//      CpriText(hnd, " %dmm", (int) (height * 4.23333 / 100 + 0.5));
      CpriCommand(hnd, CPRI_MOVE_TO_NEXT_COLUMN);
      CpriText(hnd, ptxt[pt_text_frame_height2]);
//      CpriText(hnd, " %d%%", height);
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCommand(hnd, CPRI_END_GRID);

      /* barcode test */
      CpriCommand(hnd, CPRI_LINE_HEIGHT, 100);
//      CpriText(hnd, "%s:", ptxt[pt_bar_code_test]);
      CpriCommand(hnd, CPRI_BAR_CODE, "12345678");
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCommand(hnd, CPRI_NEW_LINE);

      /* typeface test */
      CpriText(hnd, ptxt[pt_typefaces_in_order]);
      CpriText(hnd, ptxt[pt_for_each_face]);
      CpriCommand(hnd, CPRI_LINE_HEIGHT, 120);
      CpriCommand(hnd, CPRI_NEW_LINE);
      names = CpriGetTypefaceNames();
      for (i = 0; i < NOTYPEFACES; i++) {
         CpriCommand(hnd, CPRI_FONT_TYPEFACE, TIMES, 0, 0);
         CpriCommand(hnd, CPRI_FONT_SIZE, 8);

         CpriText(hnd, names[i]);
         CpriText(hnd, " ");
         for (j = 0; j < 2; j++) {
            for (k = 0; k < 2; k++) {
               CpriCommand(hnd, CPRI_FONT_TYPEFACE, i, k, j);
               CpriCommand(hnd, CPRI_FONT_SIZE, 12);
               CpriText(hnd, " GRTQWMJjxga");
            }
         }
         CpriCommand(hnd, CPRI_NEW_LINE);
      }

      /* underlined text */
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCommand(hnd, CPRI_LINE_HEIGHT, 100);
      CpriCommand(hnd, CPRI_FONT_TYPEFACE, TIMES, 0, 0);
      CpriCommand(hnd, CPRI_UNDERLINE, 1);
      CpriText(hnd, ptxt[pt_underline_text]);
      CpriCommand(hnd, CPRI_UNDERLINE, 0);
      /* scalable font text */
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCommand(hnd, CPRI_NEW_LINE);
      for (i = 0; i < 5; i++) {
         CpriCommand(hnd, CPRI_FONT_SIZE, sizes[i]);
         CpriText(hnd, ptxt[pt_fontsizes]);
//         CpriText(hnd, "  %s %d", ptxt[pt_fontsizes], sizes[i]);
      }
      CpriCommand(hnd, CPRI_NEW_LINE);
      for (; i < 8; i++) {
         CpriCommand(hnd, CPRI_FONT_SIZE, sizes[i]);
//         CpriText(hnd, "  %s %d", ptxt[pt_fontsizes], sizes[i]);
      }
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCommand(hnd, CPRI_FONT_SIZE, sizes[i]);
      CpriText(hnd, ptxt[pt_fontsizes]);
//      CpriText(hnd, "  %s %d", ptxt[pt_fontsizes], sizes[i]);
      CpriCommand(hnd, CPRI_NEW_LINE);
      CpriCloseEntry(hnd);
   }
}

static char *MoveToNextCell(t_driver *dd, char *p)
{
   t_tblstate *ts;
   t_tab_frame *tf;
   int ymove;
   
   ts = dd->tstate;
   tf = &ts->tf;
   if (ts->ul && tf->curtab == 0)
      p += sprintf(p, HP_UNDERLINE_OFF);
   if (tf->curtab == 0) {
      p += sprintf(p, HP_ABSMOVEX, tf->tabs[tf->curtab].scaled);
      /* The current height should be used by this new row */
      ymove = SetLineHeight(dd); /* Returns baseline to baseline measurement */
      if (dd->ypos + ymove + BASELINE_TO_CELLBOTTOM(ts) > dd->tablebottom) {
         p = NewPage(dd, p);
         dd->tstate->newtable = 1;
         if (dd->current_play_macro)
            return p; /* If macro playing then we have to do the rest after the macro */
      } else {
         if (ts->newtable) {
            dd->ypos = dd->tabletop + CELLTOP_TO_BASELINE(ts);
         } else {
            dd->ypos += ymove;
         }
         p += sprintf(p, HP_ABSMOVEY, dd->ypos);
      }
      if (!dd->gridinited && ts->gridon) {
         p = DrawUpperGridSegment(dd, p);
         dd->gridinited = 1;
      }
      ts->newtable = 0;
   }
   /* Adjust the x-pos a little (1mm) to not touch the grid */
   p += sprintf(p, HP_ABSMOVEX, tf->tabs[tf->curtab].scaled + 12);
   if (ts->ul && tf->curtab == 0)
      p += sprintf(p, HP_UNDERLINE_ON);
   _CpriMoveToNextCell(tf);
   return p;
}

static void ResetDriverState(t_driver *dd)
{
   int xt_fonts,i;
   struct t_printerq *q;
   
   xt_fonts = dd->xt_fonts;
   q = dd->q;
   Release(dd->outmembuf);
   _CpriDestroyTabFrame(&dd->tstate->tf);
   if (dd->macros) {
      for (i=0; i<dd->nmacros; i++)
          DestroyMacro(dd->macros+i);
      Release(dd->macros);
      dd->macros = NULL;
   }
   dd->nmacros = 0;
   dd->current_record_macro = NULL;
   dd->macroids = 0;
   memset(dd, 0, sizeof(t_driver));
   InitDriverData(dd);
   dd->q = q;
   dd->xt_fonts = xt_fonts;
}

static char *PrintCheck(t_driver *dd nouse, char *p)
{
   p +=sprintf(p, "\x0e\373\x0f");
   return p;
}

static char *SetFontSize(t_driver *dd, char *p)
{
   p +=sprintf(p, HP_SIZE, dd->tstate->size);
   return p;
}

static char *ManualFeed(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_MANUALFEED);
   return p;
}

static char *CasetteFeed(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_CASSETTEFEED);
   return p;
}

static char *Pushpos(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_PUSHPOS);
   return p;
}

static char *Poppos(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_POPPOS);
   return p;
}

static char *StartRecordHPMacro(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_STARTMACRO);
   return p;
}

static char *StopRecordHPMacro(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_STOPMACRO);
   return p;
}

static char *PlayHPMacro(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_RUNMACRO);
   return p;
}

static char *DeleteHPMacros(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_DELETEMACROS);
   return p;
}

static char *SetHPMacroId(t_driver *dd nouse, char *p, int id)
{
   p += sprintf(p, HP_MACROID, id);
   return p;
}

static char *SetLandscape(t_driver *dd, char *p)
{
   dd->paper_height = A4_PAGE_WIDTH;
   p += sprintf(p, "%s%s", HP_LANDSCAPE, HP_MARGINS);
   return p;
}

static char *SetPortrait(t_driver *dd, char *p)
{
   dd->paper_height = A4_PAGE_HEIGHT;
   p += sprintf(p, "%s%s", HP_PORTRAIT, HP_MARGINS);
   return p;
}

static char *SetUnderlineState(t_driver *dd, char *p)
{
   if (dd->tstate->ul)
      p += sprintf(p, HP_UNDERLINE_ON);
   else
      p += sprintf(p, HP_UNDERLINE_OFF);
   return p;
}

/* clear all stuff, send printable area, default symbols(windows) for primary
   set and default symbols(PC8) and fix-spacing for secondary set */
static char *HpReset(t_driver *dd nouse, char *p)
{
   p += sprintf(p, HP_UNDERLINE_OFF);
   p += sprintf(p, "%s%s%s%s\33)s0P\33)s10H\33)10U", HP_DELETEMACROS,
                     HP_MARGINS, HP_LINES_PER_INCH, HP_SYMBOLSET);
   return p;
}

static void ProcessComponent(t_driver *dd)
{
   char s[1000], tmp[100];
   int c, c2;
   int param, w, barw, barh, pattern, grey, embedded;
   double fwidth, fheight, fpar;
   struct t_printerq *q;
   t_tab_frame *tf;

   q = dd->q;
   tf = &dd->tstate->tf;
   c = _CpriGetChar(q);
   switch (c) {
   case 27:
      c = _CpriGetChar(q);
      switch (c) {
      case 27: /* Undocumented test mode to make it possible for an
                  application to pass PCL escape sequences directly
                  to the printer, without interfereing with the
                  CPRI escape sequenses. Must be terminated by two
                  escape codes in sequence (not sent to the printer). */
         c2 = 0;
         do {
            c = _CpriGetChar(q);
            if (c == 27) {
               c2 = _CpriGetChar(q);
               if (c2==27) {
                  *dd->outend = 0;
               } else {
                  *dd->outend++ = c;
                  *dd->outend++ = c2;
               }
            } else
               *dd->outend++ = c;
         } while (c2!=27);
         break;
      case EPC_NEW_LINE:
         dd->outend = NewLine(dd, dd->outend);
         break;
      case EPC_ABS_MOVEY:
         _CpriGetIntParam(q, &param);
         dd->ypos = param*dd->scaling + 0.5;
         dd->outend += sprintf(dd->outend, HP_ABSMOVEY, dd->ypos);
         break;
      case EPC_REL_MOVEX:
         _CpriGetFloatParam(q, &fpar);
         param = fpar * dd->scaling + 0.5;
         dd->outend += sprintf(dd->outend, HP_RELMOVEX, param);
         break;
      case EPC_REL_MOVEY:
         _CpriGetIntParam(q, &param);
         param = param * dd->scaling + 0.5;
         dd->outend += sprintf(dd->outend, HP_RELMOVEY, param);
         dd->ypos += param;
         break;
      case EPC_ABS_MOVEX:
         _CpriGetIntParam(q, &param);
         param = (param - 5) * dd->scaling + 0.5;
         dd->outend += sprintf(dd->outend, HP_ABSMOVEX, param);
         break;
      case EPC_RECTANGLE:
         _CpriGetFloatParam(q, &fpar);
         fwidth = fpar * dd->scaling;
         _CpriGetFloatParam(q, &fpar);
         fheight = fpar * dd->scaling;
         dd->outend = DrawRectAtCurrentPos(dd->outend, fwidth, fheight);
         break;
      case EPC_NR_OF_COPIES:
         if (!dd->copies)
            _CpriGetIntParam(q, &dd->nocopies);
         else
            _CpriGetIntParam(q, &param);
         break;
      case EPC_BEGIN_TO_BE_REPEATED:
         StartRecordMacro(dd);
         break;
      case EPC_END_TO_BE_REPEATED:
         dd->outend = StopRecordMacro(dd, dd->outend);
         break;
      case EPC_TABLE:
         _CpriGetIntParam(q, &param);
         param -= 5; /* due to stupid hp-machine */
         if (param < 0)
            param = 0;
         _CpriInitTabs(tf, param, dd->scaling);
         
         _CpriGetIntParam(q, &dd->tabletop);
         if (dd->tabletop < 6)
            dd->tabletop = 6;
         dd->tabletop = dd->tabletop * dd->scaling + 0.5;
         
         _CpriGetIntParam(q, &param);
         if (param < 6)
            param = 6;
         dd->tablebottom = dd->paper_height - param * dd->scaling + 0.5;
         dd->outend += sprintf(dd->outend, HP_ABSMOVEX, tf->tabs[0].scaled);
         if (dd->ypos < dd->tabletop) {
            dd->ypos = dd->tabletop;
            dd->outend += sprintf(dd->outend, HP_ABSMOVEY, dd->ypos);
         }
         dd->tstate->newtable = 1;
         break;
      case EPC_END_TABLE:
         _CpriDestroyTabFrame(tf);
         break;
      case EPC_TEXT_TO_VLINE_SPACE:
         _CpriGetIntParam(q, &barw);
         break;
      case EPC_TEXT_TO_HLINE_SPACE:
         _CpriGetIntParam(q, &dd->tstate->linepospercent);
         dd->tstate->text2grid = dd->tstate->rowheight * dd->tstate->linepospercent / 100.0 + 0.5;
         break;
      case EPC_HORIZONTAL_BAR:
         _CpriGetIntParam(q, &barw);
         _CpriGetIntParam(q, &barh);
         _CpriGetIntParam(q, &pattern);
         _CpriGetIntParam(q, &grey);
         _CpriGetIntParam(q, &embedded);
         dd->outend = MakeDiagramBar(dd, dd->outend, barw, barh, pattern, grey, embedded);
         break;
      case EPC_END_BAR_LABEL:
         if (dd->printing_embedded) {
            dd->printing_embedded = 0;
            /* This passes a signal in the recursive return */
            return;
         }
         break;
      case EPC_CREATE_COLUMN:          /* set the width of a new column */
         _CpriGetIntParam(q, &param);
         _CpriSetTab(tf, param, dd->scaling);
         break;
      case EPC_ACTIVATE_COLUMNS:
         _CpriActivateTabs(tf);
         break;
      case EPC_DEACTIVATE_COLUMN:
         _CpriGetIntParam(q, &param);
         _CpriRemovePriTab(tf, param);
         break;
      case EPC_MOVE_TO_NEXT_COLUMN:
          /* Moves the print position and logical align borders to next cell */
         if (tf->curtab < tf->n)
            dd->outend = MoveToNextCell(dd, dd->outend);
         break;
      case EPC_RESET:
         dd->outend = HpReset(dd, dd->outend);
         break;
      case EPC_REMOVE_COLUMNS:       /* clear tabs */
         ClearTabs(dd);
         break;
      case EPC_MERGE_ALL_COLUMNS:
         _CpriMergeAllColumns(tf);
         break;
      case EPC_START_GRID:   /* set auto grid */
         if (!dd->tstate->gridon) {
            dd->tstate->gridon = 1;
            dd->gridinited = 0;
         }
         break;
      case EPC_END_GRID:   /* finish auto grid */
         dd->tstate->gridon = 0;
         break;
      case EPC_SCALING:
         _CpriGetIntParam(q, &param);
         if (param == 0)
            dd->scaling = HP_SCALING / 10;   /* 30 */
         else
            dd->scaling = HP_SCALING / 25.4; /* 11.8 */
         break;
      case EPC_LINE_HEIGHT:
         _CpriGetIntParam(q, &dd->tstate->hightpercent);
         if (dd->tstate->hightpercent == 0)
            dd->tstate->hightpercent = 100;
         break;
      case EPC_FIX_ROW_HEIGHT:
         _CpriGetIntParam(q, &param);
         dd->tstate->fixrowheight = param * dd->scaling + 0.5;
         break;
      case EPC_NEW_PAGE:
         dd->outend = NewPage(dd, dd->outend);
         break;
      case EPC_DOSTEXT:
         _CpriGetIntParam(q, &dd->dos);
         break;
      case EPC_ALIGN:
         _CpriGetIntParam(q, &param);
         _CpriGetStringParam(q, s);
         w = string_width(s, dd->tstate->face, dd->tstate->stroke, dd->tstate->style, dd->tstate->size, dd->xt_fonts, HP_SCALING);
         switch (param) {
         case ALIGN_LEFT:
            if (dd->postponed_superscript) {
               GetPostponedSuperscript(dd, tmp);
               dd->outend = RightAlignSuperscriptToCurrentPosition(dd, dd->outend, tmp);
            }
            dd->outend += sprintf(dd->outend, s);            
            break;
         case ALIGN_CENTER:
            dd->outend = Center(dd, dd->outend, s, w);
            break;
         case ALIGN_RIGHT:
            dd->outend = AlignToRightCellEdge(dd, dd->outend, s, w);
            break;
         case ALIGN_RIGHTHE:
            dd->outend = AlignHalfToRightCellEdge(dd, dd->outend, s, w);
            break;
         case ALIGN_RIGHTDP:
            dd->outend = AlignDeciToRightCellEdge(dd, dd->outend, s, w);
            break;
         }
         break;
      case EPC_SUP_LALIGN: /* Left align to previous text/left edge */
      case EPC_SUP:
         _CpriGetStringParam(q, s);
         dd->outend = SuperscriptAtCurrentPos(dd, dd->outend, s, dd->tstate->size);
         break;
      case EPC_SUP_RALIGN: /* Right align to subsequent text */
         dd->postponed_superscript = _CpriGetInputPos(q);
         _CpriGetStringParam(q, s);
         break;
      case EPC_SUP_RALIGN_CELL:
         _CpriGetStringParam(q, s);
         dd->outend = SuperscriptRightAligned(dd, dd->outend, s);
         break;
      case EPC_STRIKE:
         _CpriGetStringParam(q, s);
         dd->outend = MakeStrikeText(dd, dd->outend, s);
         break;
      case EPC_HTML_TAG:      /* html-tag - clear input and ignore */
         _CpriGetStringParam(q, s);
         break;
      case EPC_END_HTML_TAG:   /* html-tagend  - ignore */
      case EPC_END_START_SEQUENCE:
         break;
      case EPC_BAR_CODE:
         _CpriGetStringParam(q, s);
         dd->outend = PrintBarCodeAtCurPos(dd, dd->outend, s);
         break;
      case EPC_BAR_CODE_W:
         _CpriGetIntParam(q, &param);
         SetBarCodeWidth(dd, param);
         break;
      case EPC_BAR_CODE_H:
         _CpriGetIntParam(q, &param);
         SetBarCodeHeight(dd, param);
         break;
      case EPC_CHECK:
         dd->outend = PrintCheck(dd, dd->outend);
         break;
      case EPC_FONT_TYPEFACE:
         dd->outend = SetTypface(dd, dd->outend);
         break;
      case EPC_FONT_SIZE:
         _CpriGetIntParam(q, &dd->tstate->size);
         dd->outend = SetFontSize(dd, dd->outend);
         break;
      case EPC_UNDERLINE:
         _CpriGetIntParam(q, &dd->tstate->ul);
         dd->outend = SetUnderlineState(dd, dd->outend);
         break;
      case EPC_PORTRAIT:
         dd->outend = SetPortrait(dd, dd->outend);
         break;
      case EPC_LANDSCAPE:
         dd->outend = SetLandscape(dd, dd->outend);
         break;
      case EPC_MANUAL_FEED:
         dd->outend = ManualFeed(dd, dd->outend);
         break;
      case EPC_CASSETTE_FEED:
         dd->outend = CasetteFeed(dd, dd->outend);
         break;
      case EPC_PUSH_POS:
         dd->outend = Pushpos(dd, dd->outend);
         break;
      case EPC_POP_POS:
         dd->outend = Poppos(dd, dd->outend);
         break;
      case EPC_START_RECORD_MACRO:
         dd->outend = StartRecordHPMacro(dd, dd->outend);
         break;
      case EPC_STOP_RECORD_MACRO:
         dd->outend = StopRecordHPMacro(dd, dd->outend);
         break;
      case EPC_EXECUTE_MACRO:
         dd->outend = PlayHPMacro(dd, dd->outend);
         break;
      case EPC_MACROID:
         _CpriGetIntParam(q, &param);
         dd->outend = SetHPMacroId(dd, dd->outend, param);
         break;
      case EPC_DELETE_MACROS:
         dd->outend = DeleteHPMacros(dd, dd->outend);
         break;
      }
      break;
   case '\14':
      *dd->outend++ = c;
      *dd->outend = _CpriGetChar(q);
      if (*dd->outend++=='\15') {
         *dd->outend++ = '\033';
         *dd->outend++ = 'E';
      }
      *dd->outend = 0;
      dd->ready = 1;
      break;
   case EOF:
      *dd->outend++ = EOF;
      *dd->outend = 0;
      dd->ready = 1;
      return;
   default:
      if (dd->printing_embedded)
         dd->embedded_string = dd->outend;
      if (dd->postponed_superscript) {
         GetPostponedSuperscript(dd, tmp);
         dd->outend = RightAlignSuperscriptToCurrentPosition(dd, dd->outend, tmp);
      }
      *dd->outend++ = c;
      _CpriGetStringParam(q, s);
      _CpriUnGetChar(q);
      dd->outend += sprintf(dd->outend, s);
   }
   /* If the current in-sequence produced no output to feed the printer (i.e.
      it was an internal handled command) then we have to look for next */
   if (*dd->outptr==0)
      ProcessComponent(dd);
}

static int HPDriver(t_driver *dd)
{
   if (*dd->outptr)
      return (unsigned)(unsigned char)*dd->outptr++;
   if (dd->ready) {
      if (--dd->nocopies > 0) {
         InitDriverData(dd);
         _CpriSetInputPos(dd->q, dd->instart);
         dd->copies = 1;
      } else {
         dd->copies = 0;
         ResetDriverState(dd); /* Clears all pointers */
         /* This forces the reader to see end-of-file and fetch a new
            q-element if there is any. */
         _CpriGetChar(dd->q);
         _CpriUnGetChar(dd->q);
      }
   }
   if (dd->instart == NULL) {
      _CpriInitBufferedInput(dd->q);
      dd->instart = _CpriGetInputPos(dd->q);
      dd->ready = 0;
   }
   dd->outend = dd->outptr = dd->outmembuf;
   *dd->outend = 0;
   ProcessComponent(dd);
   if (*dd->outptr)
      return (unsigned)(unsigned char)*dd->outptr++;
   /* No more to process but it may be multiple copies */
   return EOF;
}

static t_cpri_vtable hp_vtable = {HPInit, HPClose, HPDriver};
static t_cpri_vtable kyocera_vtable = {KyoceraInit, HPClose, HPDriver};
static t_cpri_vtable brother_vtable = {BrotherInit, HPClose, HPDriver};
t_cpri_vtable *_cpri_hp_vtable = &hp_vtable;
t_cpri_vtable *_cpri_nec_vtable = &hp_vtable;
t_cpri_vtable *_cpri_kyocera_vtable = &kyocera_vtable;
t_cpri_vtable *_cpri_brother_vtable = &brother_vtable;

#endif
