/* Module DRV_HTML.C
   Contains the print-driver for creating html-files */

#include <string.h>
#include <allegro.h>

#include <cgui.h>
#include <cgui/mem.h>

#include "spooler.h"
#include "drv_html.h"
#include "cpri.h"
#include "pridrive.h"
#include "drv_common.h"
#include "labldata.h"
#include "dialpri.h"
#include "cpri.ht"

#define MAXLINKLEN 10000

typedef struct t_driver {
   t_tab_frame tf;              /* A list of the current tabs (in three
                                   different scalings) */
   float scaling;               /* Scaling factor (currently values will be
                                   scaled from text-box or mm to suitabble
                                   values for the printer) */
   int gridon;                  /* It is necessary to set gridon after
                                   defining the set of tabs to be used */
   int w;                       /* Width of the current cell (or cells in case
                                   of multi column spanning cells) */
   int span;                    /* The number of columns the current cell spanns */
   t_tab margin;
   struct t_printerq *q;
} t_driver;

static void GetHTMLLink(struct t_printerq *q, char *link)
{
   link += sprintf(link, "<a href=\"");
   _CpriGetStringParam(q, link);
   strcat(link, "\">");
}

static void HtmlFileHeader(FILE *dst, const char *title)
{
   fprintf(dst,
              "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
              "<html>\n"
              "<head>\n"
              "   <meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">"
              "   <title>");
   fprintf(dst, title);
   fprintf(dst, "</title>\n" "</head>\n" "<body>\n");
}

static void Tab(t_driver *dd, int bold, int ul, int italic, char *celltext,
                   char *line, int align, int size, char **faces, int face)
{
   char *p;
   char *align_texts[] = {
      "",                       /* left is default */
      "align=center",
      "align=right",
      "align=right",
      "align=right"
   };

   p = line + strlen(line);
   p += sprintf(p, "  <td width=\"%d\"", dd->w);
   if (dd->span>1)
      p += sprintf(p, " colspan=%d", dd->span);
   if (align)
      p += sprintf(p, " %s", align_texts[align]);
   p += sprintf(p, ">");
   if (face)
      p += sprintf(p, "<span style=\"font-size:%d; font-family:%s\">", size, faces[face]);
   else
      p += sprintf(p, "<span style=\"font-size:%d\">", size);
   if (bold)
      p += sprintf(p, "<b>");
   if (ul)
      p += sprintf(p, "<u>");
   if (italic)
      p += sprintf(p, "<i>");
   p += sprintf(p, celltext);
   if (*celltext == 0 && dd->gridon)
      p += sprintf(p, "&nbsp;");
   if (align==ALIGN_RIGHTHE) {
      if (celltext[strlen(celltext) - 1] != '\xbd')
         p += sprintf(p, "&nbsp;&nbsp;&nbsp;");
   } else if (align==ALIGN_RIGHTDP) {
      if (celltext[strlen(celltext) - 1] == '.')
         p += sprintf(p, "&nbsp;&nbsp;");
      else if (celltext[strlen(celltext) - 2] == '.')
         ;
      else
         p += sprintf(p, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
   }
   if (bold)
      p += sprintf(p, "</b>");
   if (ul)
      p += sprintf(p, "</u>");
   if (italic)
      p += sprintf(p, "</i>");
   if (face)
      p += sprintf(p, "</span>");
   p += sprintf(p, "</td>\n");
   *celltext = 0;
}

static struct t_driver *HtmlInit(struct t_printerq *q)
{
   t_driver *dd;

   dd = GetMem0(t_driver, 1);
   dd->q = q;
   dd->scaling = 4;
   return dd;
}

static void HtmlClose(t_driver *dd)
{
   if (dd)
      Release(dd);
}

/* "CELLPADDING=\"%d\"" could be used for rowheight surrounds the text in a
   cell with %d pixels of space */
static int HtmlDriver(t_driver *dd)
{
   FILE *dst;
   char line[5000] = "", tmp[5000], s[500], *fn;
   int nr, virgin = 1, italic=0, bold=0, underline=0, size=0, face=0;
   int barw,barh,pattern,gray,embedded,align=0;
   fpos_t border_filepos, end_filepos;
   int param, c;
   double ddummy;
   struct t_printerq *q;
   t_tab_frame *tf;
   char *htmlfaces[] = {
      "Albertus,Times",
      "Antique,Times",
      "'Times New Roman',Times",
      "Verdana,Arial,Helvetica",
      "'Universe condensed',Arial,Helvetica",
      "Coronet, 'Courier New',Courier",
      "Courier,'Courier New'",
      "Gothic,'Courier New',Courier",
   };
   char link[MAXLINKLEN+1], celltext[1000];
   char *colors[]={"#0000ff","#ccccff","#ccffcc","#ffcccc","#ccffff","#ffccff","#ffffcc"};
   const char *const*ptxt;

   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      return EOF;
   q = dd->q;
   tf = &dd->tf;
   fn = _CpriGetDestinationFilename(q);
   dst = fopen(fn, "w");
   if (dst) {
      do {
         c = _CpriGetChar(q);
         switch (c) {
         case 033:
            c = _CpriGetChar(q);
            switch (c) {
            case EPC_STRIKE:
               _CpriGetStringParam(q, s);
               strcat(celltext, "<strike>");
               strcat(celltext, s);
               strcat(celltext, "</strike>");
               break;
            case EPC_SUP_LALIGN:
            case EPC_SUP:
            case EPC_SUP_RALIGN:
            case EPC_SUP_RALIGN_CELL:
               _CpriGetStringParam(q, s);
               if (underline)
                  strcat(celltext, "</u>");
               strcat(celltext, "<sup><span style=\"font-size:smaller\">");
               strcat(celltext, s);
               strcat(celltext, "</sup></span>");
               if (underline)
                  strcat(celltext, "<u>");
               break;
            case EPC_FONT_TYPEFACE:  /* typefaces */
               _CpriGetIntParam(q, &face);
               _CpriGetIntParam(q, &bold);  /* get stroke weight */
               _CpriGetIntParam(q, &italic);  /* get style */
               break;
            case EPC_UNDERLINE: /* Underline off/on */
               _CpriGetIntParam(q, &underline);
               break;
            case EPC_CREATE_COLUMN:    /* settab */
               _CpriGetIntParam(q, &param);
               _CpriSetTab(tf, param, dd->scaling);
               break;
            case EPC_MOVE_TO_NEXT_COLUMN:
               if (tf->curtab >= tf->n)   /* too many tabs */
                  break;
               if (virgin) {
                  fprintf(dst, "\n<table border=");
                  fgetpos(dst, &border_filepos);
                  fprintf(dst, "%d width=\"%d\">\n", dd->gridon,
                          tf->tabs[tf->n - 1].scaled - tf->tabs[0].scaled);
                  virgin = 0;
               }
               _CpriMoveToNextCell(tf);
               if (tf->prevtab == 0) {
                  strcpy(line, " <tr>\n");
                  *celltext = 0;
               } else {
                  Tab(dd, bold, underline, italic, celltext, line, align,
                              size, htmlfaces, face);
               }
               dd->w = tf->tabs[tf->curtab].scaled - tf->tabs[tf->prevtab].scaled;
               dd->span = tf->curtab - tf->prevtab;
               break;
            case EPC_NEW_LINE:
               Tab(dd, bold, underline, italic, celltext, line, align,
                             size, htmlfaces, face);
               fprintf(dst, line);
               fprintf(dst, " </tr>\n");
               *line = 0;
               tf->curtab = 0;
               break;
            case EPC_ACTIVATE_COLUMNS:
               _CpriActivateTabs(tf);
               break;
            case EPC_DEACTIVATE_COLUMN:
               _CpriGetIntParam(q, &param);
               _CpriRemovePriTab(tf, param);
               break;
            case EPC_START_GRID: /* grid on */
               if (virgin) {
                  dd->gridon = 1;
               } else if (!dd->gridon) {
                  dd->gridon = 1;
                  fgetpos(dst, &end_filepos);
                  fsetpos(dst, &border_filepos);
                  fprintf(dst, "%d", dd->gridon);
                  fsetpos(dst, &end_filepos);
               }
               break;
            case EPC_END_GRID:
               dd->gridon = 0;
               break;
            case EPC_REMOVE_COLUMNS:  /* not implemented */
               break;
            case EPC_MERGE_ALL_COLUMNS:
               _CpriMergeAllColumns(tf);
               break;
            case EPC_FONT_SIZE:
               _CpriGetIntParam(q, &param);
               size = param;
/*               size = (param - 12) / 2;
               if (size > 1) {
                  size /= 2;
                  if (size > 3) {
                     size = 3;
                  }
               }*/
               break;
            case EPC_END_START_SEQUENCE:
               HtmlFileHeader(dst, "");
               break;
               /* unimplemented without parameters */
            case EPC_NEW_PAGE:
            case EPC_CHECK:
            case EPC_PUSH_POS:
            case EPC_POP_POS:
            case EPC_DELETE_MACROS:
            case EPC_START_RECORD_MACRO:
            case EPC_STOP_RECORD_MACRO:
            case EPC_EXECUTE_MACRO:
            case EPC_PORTRAIT:
            case EPC_LANDSCAPE:
            case EPC_RESET:
            case EPC_MANUAL_FEED:
            case EPC_CASSETTE_FEED:
               break;
               /* unimplemented commands that contains 1 parameter */
            case EPC_REL_MOVEX:
            case EPC_REL_MOVEY:
            case EPC_LINE_HEIGHT:
            case EPC_SCALING:
            case EPC_MACROID:
            case EPC_ABS_MOVEX:
            case EPC_ABS_MOVEY:
            case EPC_TEXT_TO_HLINE_SPACE:
            case EPC_TEXT_TO_VLINE_SPACE:
            case EPC_DOSTEXT:
            case EPC_NR_OF_COPIES:
            case EPC_FIX_ROW_HEIGHT:
            case EPC_BAR_CODE_W:
            case EPC_BAR_CODE_H:
               _CpriGetIntParam(q, &param);
               break;
            case EPC_TABLE:
               /* The table position is actually not interesting in html
               it will either be desirable to have it left aligned or
               centerd, but since we don't support centering of the table
               we left align it (i.e. do nothing) */
               _CpriGetIntParam(q, &param); /* x */
               _CpriInitTabs(tf, param, dd->scaling);
               _CpriGetIntParam(q, &param); /* y */
               _CpriGetIntParam(q, &param); /* bottom */
               virgin = 1;
               break;
            case EPC_END_TABLE:
               _CpriDestroyTabFrame(tf);
               fprintf(dst, "</table>\n\n");
               break;
            case EPC_BEGIN_TO_BE_REPEATED:
            case EPC_END_TO_BE_REPEATED:
               /* No pages needed on the screen */
               break;
            case EPC_HORIZONTAL_BAR: /* 5 parameters */
               _CpriGetIntParam(q, &barw);
               barw *= dd->scaling;
               _CpriGetIntParam(q, &barh);
               _CpriGetIntParam(q, &pattern);
               _CpriGetIntParam(q, &gray);
               _CpriGetIntParam(q, &embedded);
               break;
            case EPC_END_BAR_LABEL:
               strcpy(tmp, celltext);
               if (embedded)
                  sprintf(celltext, "<table width=%d bgcolor=\"%s\"><tr><td>"
                       "<span style=\"font-size:%d; font-family:%s>%s%s%s%s%s%s%s</span>"
                       "</td></tr></table>", barw, colors[pattern],
                       size, htmlfaces[face], bold?"<b>":"",underline?"<u>":"",italic?"<i>":"",
                       tmp,
                       bold?"</b>":"",underline?"</u>":"",italic?"</i>":"");
               else
                  sprintf(celltext, "%s<br><table width=%d bgcolor=\"%s\"><tr><td>"
                       "&nbsp;"
                       "</td></tr></table>", tmp, barw, colors[pattern]);
               break;
            case EPC_RECTANGLE:
               _CpriGetFloatParam(q, &ddummy);
               _CpriGetFloatParam(q, &ddummy);
               break;
            case EPC_ALIGN:
               _CpriGetIntParam(q, &align);
               _CpriGetStringParam(q, celltext + strlen(celltext));
               break;
            case EPC_BAR_CODE:
               _CpriGetStringParam(q, tmp);  /* dummy read */
               break;
            case EPC_HTML_TAG:  /* html link begin */
               GetHTMLLink(q, link);   /* always call to get rid of text from input */
               strcat(celltext, link);
               break;
            case EPC_END_HTML_TAG: /* html link end */
               strcat(celltext, "</a>");
               break;
            default:
               break;
            }
            break;
         case '\14':
         case '\15':
         case EOF:
            break;
         default:
            sprintf(tmp, "%c", c);
            strcat(celltext, tmp);
            break;
         }
      } while (c != EOF);
      fprintf(dst, "</body>\n</html>");
      fclose(dst);
   }
   return EOF;
}

t_cpri_vtable html_vtable = {HtmlInit, HtmlClose, HtmlDriver};
struct t_cpri_vtable *_cpri_html_vtable = &html_vtable;
