/* Module SPOOLER
   Contains the functions for background printing */

#include <allegro.h>
#include <string.h>
#include <stdio.h>

#include <cgui.h>
#include <cgui/mem.h>

#include "labldata.h"
#include "spooler.h"
#include "cpri.h"
#include "dialpri.h"
#include "cpri.ht"

#include "drv_stdwin.h"
#include "drv_txt.h"
#include "drv_hp.h"
#include "drv_html.h"
#include "printini.h"

#define IdleQ(q) (_CpriEmptyQ(q) && !(q)->hold)
//#define silent_test

#ifdef silent_test
#define output_char(portaddress, chr)
#else
#ifdef DJGPP
#include <dos.h>
#define output_char(portaddress, chr)  \
      outportb(portaddress, chr);      \
      outportb(portaddress+2, 0x0d);   \
      inportb(portaddress+1);          /* dummy read to ensure a dealy of 1 */\
      inportb(portaddress+1);          /* dummy read to ensure a dealy of 1 */\
      inportb(portaddress+1);          /* dummy read to ensure a dealy of 1 */\
      outportb(portaddress+2, 0x0c);
#else
#define output_char(portaddress, chr)
#define inportb(x) 0x70
#endif
#endif

struct t_handle {
   FILE *sf;                /* The spool file */
   char sfn[20];           /* The generated spooler filename. Must be
                              preserved to be able to remove the file
                              from disk when printed. */
   char *dfn;              /* Destination name (optional, empty string if
                              not used) */
   char *po_name;          /* The identification name of the print out. This
                              name can be used in a later stage when in the
                              q-monitor */
   struct t_printerq *q;   /* The q to assign the the printout to */
   int i;
};

t_handle *hnds = NULL;
int nhnds;

typedef struct t_qelem {
   long bytes;             /* Number of bytes left to print in this file.
                              Used for viewing the printing in progress */
   char sfn[20];            /* The generated spooler filename. Must be
                              preserved to be able to remove the file
                              from disk when printed. (is equal to empty
                              string in case of buffered i.d.)*/
   char *dfn;              /* Destination name html/txt files (optional, empty
                              string if not used) */
   char *po_name;          /* Name of printout. Names are used for
                              viewing the queue */
   char *buffer;           /* If memory buffered intermediate data this
                              points to that data else it is NULL. */
   struct t_printerq *q;   /* pointer refering to the q where the print out
                              is queued */
   struct t_qelem *next;   /* Next to be printed */
} t_qelem;

typedef struct t_qmonitor {
   t_printerq *q;
   int idstate;
   int idhold;
   int idlist;
   int bytes;
   int pstate;
} t_qmonitor;

typedef struct t_primonitor {
   int event;
   int n;
   int drag;                    /* tells the close-function if the drag-drop
                                   facility */
   /* is on (to avoid unregistering a flag used by the */
   /* application */
   t_qmonitor qm[TOT_NR_OF_QUEUES];
} t_primonitor;

t_printerq printerqs[TOT_NR_OF_QUEUES];

static char *MakeSpoolFilePath(char *spoolname)
{
   char *s;

   s = GetMem(char, strlen(_cpri_spool_dir) + strlen(spoolname) + 2);
   sprintf(s, "%s/%s", _cpri_spool_dir, spoolname);
   fix_filename_slashes(s);
   return s;
}

static void MoveFrontToEmptyQ(t_printerq *src, t_printerq *dest)
{
   if (_CpriEmptyQ(dest)) {
      if (src->el) {
         dest->el = src->el;
         src->el = src->el->next;
         dest->el->next = NULL;
      }
   }
}

extern void _CpriPopFront(t_printerq *q)
{
   t_qelem *el;
   t_printerq *oq;
   char *s;

   el = q->el;
   if (el == NULL)
      return;
   if (q->sf) {
      fclose(q->sf);
      q->sf = NULL;
      q->files--;
      remove(q->el->sfn);
   }
   if (el->buffer)
      Release(el->buffer);
   q->inbuf = NULL;
   q->el = el->next;
   Release(el->po_name);
   Release(el);
   el = q->el;
   if (el) {
      q->bytes = el->bytes;
   } else {                     /* no more elements in current print-q */
      oq = printerqs + QUEUE_OPTIMALPORT;
      if (oq->installed && !oq->hold) {   /* check generic q */
         el = oq->el;
         if (el) {               /* yes there are files */
            oq->el = el->next;
            oq->files--;
            q->files = 1;
            q->bytes = el->bytes;
            q->el = el;
            el->next = NULL;
            el->q = q;
         }
      } else {
         q->bytes = 0;
         q->files = 0;
      }
   }
   if (el) {           /* a new q-element exists */
      if (*el->sfn) {   /* With a file */
         s = MakeSpoolFilePath(el->sfn);
         q->sf = fopen(s, "rb");
         Release(s);
      } else if (el->buffer) {
         q->inbuf = el->buffer;
      }
   }
}

extern int _CpriGetInputSize(t_printerq *q)
{
   if (q->el)
      return q->el->bytes;
   return 0;
}

extern void _CpriInitBufferedInput(t_printerq *q)
{
   char *s;

   if (q->el) {
      if (q->el->buffer)
         q->inbuf = q->el->buffer;
      else {
         if (q->sf)
            rewind(q->sf);
         else {
            s = MakeSpoolFilePath(q->el->sfn);
            q->sf = fopen(s, "rb");
            Release(s);
         }
         if (q->sf) {
            q->inbuf = q->el->buffer = GetMem(char, q->el->bytes+1);
            fread(q->inbuf, q->el->bytes, 1, q->sf);
            q->inbuf[q->el->bytes] = 0;
         }
      }
   }
}

extern char *_CpriGetInputPos(t_printerq *q)
{
   return q->inbuf;
}

extern void _CpriSetInputPos(t_printerq *q, char *pos)
{
   q->inbuf = pos;
}

extern int _CpriGetChar(t_printerq *q)
{
   char *s;

   if (q->inbuf) {
      if (*q->inbuf==0) {
         _CpriPopFront(q);
         return _CpriGetChar(q);
      }
      return *q->inbuf++;
   } else if (q->sf) {
      q->c = getc(q->sf);
      if (q->c == EOF) {
         _CpriPopFront(q);
         return _CpriGetChar(q);
      }
      return q->c;
   } else if (q->el && q->el->sfn) {
      s = MakeSpoolFilePath(q->el->sfn);
      q->sf = fopen(s, "rb");
      Release(s);
      if (q->sf)
         return _CpriGetChar(q);
   }
   return EOF;
}

extern void _CpriUnGetChar(t_printerq *q)
{
   if (q->inbuf)
      q->inbuf--;
   else if (q->sf)
      ungetc(q->c, q->sf);
}

/* Read an int value from q-file if no buffer is set else extracts it
   from the buffer */
extern void _CpriGetIntParam(t_printerq *q, int *val)
{
   int c, t, sgn = 1;

   t = 0;
   do {
      c = _CpriGetChar(q);
      if (c == '-') {
         sgn = -1;
         c = _CpriGetChar(q);
         break;
      }
   } while (c == ' ');
   while (c != EOF && c != 033) {
      t *= 10;
      t += c - 48;
      c = _CpriGetChar(q);
   }
   *val = t * sgn;
}

/* Read a float value from q-file if no buffer is set else extracts it
   from the buffer */
extern void _CpriGetFloatParam(t_printerq *q, double *val)
{
   char s[100], *p;
   int i = 0, c;

   p = s;
   do {
      *p++ = c = _CpriGetChar(q);
      i++;
   } while (c != EOF && c != 033 && i<99);
   *p = 0;
   sscanf(s, "%lf", val);
}

/* Read a plain text from q-file if no buffer is set else extracts it
   from the buffer */
extern void _CpriGetStringParam(t_printerq *q, char *s)
{
   int c;

   do {
      *s++ = c = _CpriGetChar(q);
   } while (c != EOF && c != 033);
   s--;
   *s = 0;
}

static void PrinterState(t_printerq *q);  /* forward dekl. */

extern char *_CpriProcessedName(t_printerq *q)
{
   if (q->el)
      return q->el->po_name;
   else
      return "";
}

static void UpdateQ(t_qmonitor *qmon)
{
   t_printerq *q;

   q = qmon->q;
   PrinterState(q);
   if (qmon->pstate != q->pstate) {
      qmon->pstate = q->pstate;
      Refresh(qmon->idstate);
   }
   if (qmon->bytes != q->bytes) {
      qmon->bytes = q->bytes;
   }
   Refresh(qmon->idlist);
}

static void UpdateMonitorList(t_printerq *q)
{
}

static FILE *CreateNextFile(t_printerq *q, char *name)
{
   FILE *fil;
   char *s;
   int i = 0;

   do {
      sprintf(name, "%dtmp%04d.spo", q->qno, q->high);
      s = MakeSpoolFilePath(name);
      fil = fopen(s, "w");
      q->high++;
      i++;
      Release(s);
   } while(fil == NULL && i<100); /* Make a reasonable number of retries */
   return fil;
}

extern void _CpriFlushPQ(t_printerq *q)
{
   t_qelem *el, *p;
   char *s;

   /* Reset the queue header data */
   if (q->sf) {
      fclose(q->sf);
      q->sf = NULL;
   }
   q->bytes = 0;
   q->files = 0;

   /* Flush files in pending print queue */
   el = q->el;
   while (el) {
      if (*el->sfn) {
         s = MakeSpoolFilePath(el->sfn);
         remove(s);
         Release(s);
      }
      if (el->buffer)
         Release(el->buffer);
      p = el->next;
      Release(el->po_name);
      Release(el);
      el = p;
   }
   q->el = NULL;
}

static void PrinterState(t_printerq *q)
{
   unsigned char status;

   if (q->qno >= 3) {
      q->busy = 0;
      q->prevstate = q->pstate = CPR_AVAILABLE;
      return;
   }
   status = inportb(q->portaddress + 1) & 0xfc;
   q->busy = 1;
   if ((status & 0x70) == 0x50) {
      if (status & 0x80) {      /* printer is ready to receive a charcter */
         q->busy = 0;
      }
      q->prevstate = q->pstate = CPR_AVAILABLE;
   } else if ((status & 0x70) != 0x10) {  /* this is an error! */
      switch ((status & 0xf0) >> 4) {
      case 8:
         q->pstate = CPR_PRIOFF;
         break;
      case 7:
         q->pstate = CPR_NC;
         break;
      case 6:
         q->pstate = CPR_OOP;
         break;
      case 4:
         q->pstate = CPR_OFFLINE;
         break;
      default:
         q->pstate = CPR_RANDOM;
         break;
      }
   } else {
      ;                         /* should never occur */
   }
}

static t_handle CreateHandle(t_printerq *q, const char *po_name, const char *dfn)
{
   FILE *fil;
   t_handle hnd;

   nhnds++;
   hnds = ResizeMem(t_handle, hnds, nhnds);
   hnd = hnds[nhnds-1] = GetMem0(struct t_handle, 1);
   fil = CreateNextFile(q, hnd->sfn);
   if (fil == NULL) {
      Release(hnd);
      return NULL;
   }
   hnd->dfn = MkString(dfn);
   hnd->po_name = MkString(po_name);
   hnd->q = q;
   hnd->i = nhnds - 1;
   hnd->sf = fil;
   return hnd;
}

static void DestroyHandle(t_handle hnd)
{
   hnds[hnd->i] = hnds[nhnds-1];
   hnds[hnd->i]->i = hnd->i;
   nhnds--;
   Release(hnd);
}


extern t_handle CpriOpenEntry(int qno, const char *print_out_name, int options nouse, const char *dfn)
{
   static char s[200];
   const char * const *ptxt;
   char *ext, buf[100];
   t_printerq *q;
   int nr;
   t_handle hnd;

#ifdef ALLEGRO_WINDOWS
   if (qno < 5)
      qno = 0;
#endif
   if (dfn && *dfn) {
      ext = get_extension(dfn);
      if (strlen(ext) < 5) {
         strcpy(buf, ext);
         strupr(buf);
         if (strcmp(buf, "HTM") == 0 || stricmp(buf, "HTML") == 0)
            qno = QUEUE_HTML_FILE;
         else
            qno = QUEUE_TEXT_FILE;
      }
   }
   CpriInitPrinter();
   q = printerqs + qno;
   if (!q->installed && (dfn==NULL || *dfn==0)) {
      ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
      if (nr < SIZE_dialpri)
        return NULL;
      strcpy(s, ptxt[pt_printout]);
      strcat(s, ptxt[pt_selq + qno]);
      strcat(s, ptxt[pt_noact]);
      Req("", s);
      return NULL;
   }

   if (qno == QUEUE_OPTIMALPORT && !q->installed)
      q->installed = 1;
   hnd = CreateHandle(q, print_out_name, dfn);
   if (hnd)
      fprintf(hnd->sf, CPRI_RESET);
   return hnd;
}

t_qelem *InsertQueueElement(t_qelem *link, t_qelem *el)
{
   if (link == NULL)
      return el;
   link->next = InsertQueueElement(link->next, el);
   return link;
}

extern int CpriCloseEntry(t_handle hnd)
{
   char *s;
   int idle;
   t_printerq *q;
   t_qelem *el;

   CpriInitPrinter();
   if (CpriCommand(hnd, "") == 0) /* Checks if valid handle */
      return 0;

   /* Endup the file with a formfeed */
   fprintf(hnd->sf, "\14\15");
   fclose(hnd->sf);
   q = hnd->q;

   el = GetMem0(t_qelem, 1);
   strcpy(el->sfn, hnd->sfn);
   el->po_name = hnd->po_name;
   el->dfn = hnd->dfn;
   el->q = q;
   idle = IdleQ(q);
   q->el = InsertQueueElement(q->el, el);

   /* Save filesize and number of files for info */
   s = MakeSpoolFilePath(q->el->sfn);
   el->bytes = file_size(s);
   q->files++;
   Release(s);
   if (idle)
      q->PrintNextChar(q);
   DestroyHandle(hnd);
   return 1;
}

extern int CpriCommand(t_handle hnd, t_command cmd, ...)
{
   va_list ap;

   if (hnd == NULL)
      return 0;
   if (hnd->i >= nhnds)
      return 0;
   if (hnd != hnds[hnd->i])
      return 0;
   va_start(ap, cmd);
   vfprintf(hnd->sf, cmd, ap);
   va_end(ap);
   return 1;
}

extern int CpriText(t_handle hnd, const char *string)
{
   return CpriCommand(hnd, string);
}

static void PrintNextChar(void *data)
{
   t_printerq *q = data;
   int nr;
   int qno = q->qno;
   int chr;
   char s[500];
   const char * const *ptxt;

   if (q->hold || (q->el == NULL))
      return;                   /* if q is empty or hold, stop printing */
   PrinterState(q);             /* check the state */
   if (!q->busy) {
      chr = q->vtable->Driver(q->drv);
      if (chr == EOF)
         return;
      output_char(q->portaddress, chr);
      GenEvent(q->PrintNextChar, q, 0, 0);
   } else { /* busy may also mean that an error is reported from the port */
      if (q->pstate == CPR_AVAILABLE) {
         GenEvent(q->PrintNextChar, q, 0, 0); /* try again as soon as possible */
      } else {                             /* a printer error */
         if (q->pstate != q->prevstate) {  /* but don't repeat error messages */
            ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
            if (nr < SIZE_dialpri)
               return;
            sprintf(s, ptxt[pt_priq_error], qno + 1, ptxt[pt_selq + qno], ptxt[pt_pstatusl + q->pstate]);
            switch (Req("", s)) {
            case 0:
               _CpriFlushPQ(q);
               break;
            case 1:
               /* protect from subsequent error messages to the operator */
               q->prevstate = q->pstate;
               break;
            case 2:
               q->hold = 1;
               break;
            }
         }  /* otherwise we already notified the operator the first time */
         /* Try again (if operator chosed some other we stop in next round)*/
         GenEvent(q->PrintNextChar, q, 1, 0);
      }
   }
}

static void PrnPrintNextChar(void *data)
{
   t_printerq *q = data;
   static FILE *prn = NULL;
   int c;

   if (prn==NULL)
      prn = fopen("PRN", "w");
   if (prn == NULL)
      return;
   c = q->vtable->Driver(q->drv);
   if (c==EOF)
      return;
   GenEvent(q->PrintNextChar, q, 0, 0);
}

static void NetPrintNextChar(void *data)
{
   t_printerq *q = data;
   static int net_cnt;
   FILE *netspoolfile;
   const char * const *ptxt;
   int nr, c;
   char *s, *tmpfil;

   tmpfil = GetMem(char, strlen(_cpri_spool_dir)+50);
   sprintf(tmpfil, "%s\\%d.net", _cpri_spool_dir, net_cnt++);
   netspoolfile = fopen(tmpfil, "w");
   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      return;
   if (netspoolfile) {
      s = GetMem(char, strlen(q->net1) + strlen(tmpfil) + strlen(q->net2) + 4);
      sprintf(s, "%s %s %s", q->net1, tmpfil, q->net2);
      do {
         c = q->vtable->Driver(q->drv);
         if (c != EOF)
            fputc(c, netspoolfile);
      } while (c!=EOF);
      fclose(netspoolfile);
      if (system(s)) {
         if (errno == ENOMEM)
            Req("", ptxt[pt_rsyserr]);
      }
      Release(s);
   } else {
      Req("", ptxt[pt_rsyserr]);
   }
   Release(tmpfil);
}


extern char *_CpriGetDestinationFilename(t_printerq *q)
{
   if (q->el && q->el->dfn)
      return q->el->dfn;
   return "";
}

static void RequestPrintNextChar(void *data)
{
   t_printerq *q = data;
   int i, nr, qno;
   static char s[200];
   const char * const *ptxt;

   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      return;
   strcpy(s, ptxt[pt_what_printer]);
   for (i = 0; i < NR_OF_QUEUES - 1; i++) {
      strcat(s, "|");
      strcat(s, ptxt[pt_selq + i]);
   }
   qno = Req("", s);
   q = printerqs + qno;
   q->PrintNextChar(q);
}

static void FilePrintNextChar(void *data)
{
   t_printerq *q = data;

   q->vtable->Driver(q->drv);
}

static void OptimalPrintNextChar(void *data)
{
   t_printerq *thisq = data;
   t_printerq *q;
   int i;

   for (i = 0; i < 3; i++) {
      q = printerqs + i;
      if (q->installed) {
         if (IdleQ(q)) {
            MoveFrontToEmptyQ(thisq, q);
            q->PrintNextChar(q);
            return;
         }
      }
   }
}

static struct t_driver *DefaultInit(struct t_printerq *data nouse) {return NULL;}
static void DefaultClose(struct t_driver *data nouse) {}

static int DefaultDriver(struct t_driver *data nouse) {return EOF;}

extern void _CpriSetPrinterDriver(t_printerq *q)
{
   static struct t_cpri_vtable default_vtable = {DefaultInit, DefaultClose, DefaultDriver};
   switch (q->qno) {
   case QUEUE_HTML_FILE:
      q->vtable = _cpri_html_vtable;
      break;
   case QUEUE_TEXT_FILE:
      q->vtable = _cpri_text_vtable;
      break;
#ifdef ALLEGRO_DOS
   case QUEUE_OPTIMALPORT:
      q->vtable = &default_vtable;
      break;
   case QUEUE_REQUEST:
      q->vtable = &default_vtable;
      break;
   default:
      if (q->qno < 3)               /* physical port */
         dosmemget(0x400 + 8 + q->qno * 2, 2, &q->portaddress);
      switch (q->brand) {
      case BRAND_HP:
         q->vtable = _cpri_hp_vtable;
         break;
      case BRAND_NEC:
         q->vtable = _cpri_nec_vtable;
         break;
      case BRAND_KYOCERA:
         q->vtable = _cpri_kyocera_vtable;
         break;
      case BRAND_BROTHER:
         q->vtable = _cpri_brother_vtable;
         break;
      default: /* BRAND_UNKNOWN */
         q->vtable = _cpri_unkown_vtable;
         break;
      }
#elif defined ALLEGRO_WINDOWS
   default:
      q->vtable = _cpri_stdwin_vtable;
#else
   default:
      q->vtable = _cpri_html_vtable;
#endif
      break;
   }

#ifdef DJGPP
   switch (q->qno) {
   case QUEUE_HTML_FILE:
      q->PrintNextChar = FilePrintNextChar;
      break;
   case QUEUE_TEXT_FILE:
      q->PrintNextChar = FilePrintNextChar;
      break;
   case QUEUE_DOSPRINT:
      q->PrintNextChar = PrnPrintNextChar;
      break;
   case QUEUE_OPTIMALPORT:
      q->PrintNextChar = OptimalPrintNextChar;
      break;
   case QUEUE_REQUEST:
      q->PrintNextChar = RequestPrintNextChar;
      break;
   case QUEUE_NETPORT:
      q->PrintNextChar = NetPrintNextChar;
      break;
   default:
      q->PrintNextChar = PrintNextChar;
      break;
   }
#else
   q->PrintNextChar = FilePrintNextChar;
#endif
   q->drv = q->vtable->Init(q);
}

static void CyclicUpdatePrinterMon(void *data)
{
   t_primonitor *primon = data;
   int i;

   for (i = 0; i < primon->n; i++)
      UpdateQ(primon->qm + i);
   GenEvent(CyclicUpdatePrinterMon, primon, 1000, primon->event);
}

static void Flush(void *data)
{
   _CpriFlushPQ(data);
   UpdateMonitorList(data);
}

static void ClosePrinterWin(void *data)
{
   t_primonitor *primon = data;


   CloseWin(NULL);
   if (primon->drag)
      UnRegisterDragFlag(0x80000000);
   Release(primon);
}

static void HoldHandler(void *data)
{
   t_printerq *q = data;

   if (!q->hold) {               /* run-command */
      if (_CpriEmptyQ(q) && q->qno < 3)
         MoveFrontToEmptyQ(printerqs + QUEUE_OPTIMALPORT, q);
      q->PrintNextChar(q);
   }
}

static void None(int id nouse, void *obj nouse)
{
}

static int LPPrinterRow(void *list, char *s)
{
   t_qelem *el = list;
   t_printerq *q;

   if (el) {
      q = el->q;
      sprintf(s, "%s %ld", el->po_name, el->bytes);
      if (el == q->el && q->qno != QUEUE_OPTIMALPORT)
         sprintf(s + strlen(s), " %d", q->bytes);
   }
   return 0;
}

static void *LinkPrinterRow(void *head, void *cur)
{
   if (cur)
      return ((t_qelem *) cur)->next;
   else {
      return ((t_printerq *) head)->el;
   }
}

static void LinkQEntry(t_printerq *q, t_qelem *el, int x)
{
   t_qelem *p;
   int i;

   p = q->el;
   if (p == NULL) {
      el->next = NULL;
      q->el = el;
   } else {
      for (i = 0; p->next && i < x; i++, p = p->next);
      el->next = p->next;
      p->next = el;
   }
   el->q = q;
}

static void UnlinkQEntry(t_printerq *q, t_qelem *el)
{
   t_qelem *p;

   if (el == q->el)
      q->el = el->next;
   else {
      p = q->el;
      while (p && p->next) {
         if (p->next == el) {
            p->next = el->next;
            break;
         }
         p = p->next;
      }
   }
}

static void RemovePrintout(int id nouse, void *object)
{
   t_printerq *q;
   t_qelem *el = object;

   if (el) {
      q = el->q;
      UnlinkQEntry(q, el);
      q->files--;
      UpdateMonitorList(q);
   }
}

static int GripPrintOut(void *src, int reason, void *srclist nouse, int i)
{
   t_qelem *el;
   t_printerq *q;
   static int hold;
   int ok = 0;

   if (src == NULL)
      return 0;
   el = src;
   q = el->q;
   if (i || q->qno == QUEUE_OPTIMALPORT)
      ok = 1;
   switch (reason) {
   case DD_GRIPPED:
      hold = q->hold;           /* save */
      q->hold = 1;              /* hold the queue while dragging */
      break;
   case DD_SUCCESS:
      q->hold = hold;
      HoldHandler(q);
      break;
   }
   return ok;
}

static int DropPrintOut(void *dest nouse, int reason, void *src,
                        void *destlist, int i, int flags nouse)
{
   t_qelem *qe = src;
   t_printerq *sq = qe->q;
   t_printerq *q;
   static int hold;

   q = destlist;
   switch (reason) {
   case DD_OVER_DROP:
      if (i == 0)
         return 0;
      hold = q->hold;           /* save */
      q->hold = 1;              /* hold the queue while over */
      break;
   case DD_END_OVER_DROP:
      q->hold = hold;
      HoldHandler(q);
      break;
   case DD_SUCCESS:
      UnlinkQEntry(sq, qe);
      sq->files--;

      LinkQEntry(q, src, i - 1);
      q->files++;
      q->hold = hold;
      HoldHandler(q);
      break;
   }
   return 1;
}

extern void CpriPrinterQueueMonitor(void)
{
   int i, j, n, id, nr, idstate, idhold, drag;
   static const char *tmp[6];
   t_primonitor *primon;
   t_qmonitor *qmon;
   t_printerq *q;
   const char * const *ptxt;

   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      exit(-1);
   CpriInitPrinter();
   primon = GetMem0(t_primonitor, 1);
   primon->event =
       MkDialogue(ADAPTIVE, ptxt[pt_printer_contr], W_FLOATING);
   SetDistance(4, 10);
   AutoHotKeys(OFF);
   for (i = 0; i < 6; i++)
      tmp[i] = ptxt[pt_pstatus + i];
   drag = RegisterDragFlag(0x80000000);
   for (i = j = 0; i < NR_OF_QUEUES; i++) {
      if (i >= 3 && i != QUEUE_OPTIMALPORT)
         continue;
      q = printerqs + i;
      if (q && q->installed) {
         if (i < 3) {           /* only real ports may have port status */
            n = 6;
         } else {               /* optimal-q and network-q is always ready */
            n = 1;
         }
         qmon = &primon->qm[j];
         PrinterState(q);
         qmon->q = q;           /* set pointer to make it possible for the q
                                   monitor to make conditional (on state)
                                   cyclic refresh */
         qmon->bytes = q->bytes;

         StartContainer(DOWNLEFT, ADAPTIVE, ptxt[pt_selq + i], CT_BORDER);
         id = AddList(DOWNLEFT | FILLSPACE, q, &q->files, 300, 0,
                     LPPrinterRow, None, 4);
         SetLinkedList(id, LinkPrinterRow);
         if (drag) {
            SetListGrippable(id, GripPrintOut, 0x80000000, LEFT_MOUSE);
            SetListDroppable(id, DropPrintOut, 0x80000000);
         }
         SetDeleteHandler(id, RemovePrintout);

         idstate =
             AddDropDownS(DOWNLEFT | ALIGNRIGHT, 0, "", &q->pstate, tmp,
                          n);
         DeActivate(idstate);
         AddButton(LEFT, ptxt[pt_flush_q], Flush, q);
         idhold = MkRadioContainer(LEFT, &q->hold, R_HORIZONTAL);
         AddRadioButton(ptxt[pt_hold + 0]);
         AddRadioButton(ptxt[pt_hold + 1]);
         EndRadioContainer();
         AddHandler(idhold, HoldHandler, q);
         qmon->idhold = idhold;
         qmon->idstate = idstate;
         qmon->idlist = id;
         EndContainer();
         j++;
      }
   }
   AutoHotKeys(ON);
   primon->n = j;
   AddButton(DOWNLEFT, ptxt[pt_close], ClosePrinterWin, primon);
   DisplayWin();
   GenEvent(CyclicUpdatePrinterMon, primon, 1000, primon->event);
}

extern int _CpriIsPrinting(void)
{
   int i, nr, flush = 0, ask_first = 1;
   const char * const *ptxt;

   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      exit(-1);
   for (i = 0; i < TOT_NR_OF_QUEUES; i++) {
      if (printerqs[i].installed) {
         if (!_CpriEmptyQ(printerqs + i)) {
            if (ask_first)
               flush = Req("", ptxt[pt_print_progr]); /* 0 -> keep on printing */
            if (flush) {
               _CpriFlushPQ(printerqs + i);
               ask_first = 0;
            } else {
               return 1;
            }
         }
      }
   }
   return 0;
}

extern int _CpriIsPrinting2(void)
{
   int i, nr;
   const char * const *ptxt;

   ptxt = LoadCompiledTexts(cpri_labels, "dialpri", &nr);
   if (nr < SIZE_dialpri)
      exit(-1);
   for (i = 0; i < TOT_NR_OF_QUEUES; i++) {
      if (printerqs[i].installed) {
         if (!_CpriEmptyQ(printerqs + i)) {
            Req("", ptxt[pt_print_progr2]);
            return 1;
         }
      }
   }
   return 0;
}

extern int _CpriEmptyQ(struct t_printerq *q)
{
   return q->el == NULL;
}

extern void _CpriCloseQueue(t_printerq *q)
{
   if (q->installed && q->vtable)
      q->vtable->Close(q->drv);
   q->drv = NULL;
   if (!_CpriEmptyQ(q))
      _CpriFlushPQ(q);
}

